// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "CMachineWorld.h"

#include <math.h>
#include <stdio.h>

static const float CLAMP_VELOCITY_MIN = 0.0002f;
static const float CLAMP_VELOCITY_MAX = 0.02f;
static const float AIR_RESISTANCE = 0.96f;
static const float TENSION = 2.5f;

CMachineWorld::CMachineWorld() : fGravX(0), fGravY(0)
{
}

CMachineWorld::~CMachineWorld()
{
    /*! delete all existing machines */
    while(MachineList.size() > 0)
    {
        CMachine *pCurMachine = MachineList.front();

        delete pCurMachine;

        MachineList.pop_front();
    }
}

void CMachineWorld::IncrementTime(float step)
{
    std::list<CMachine *>::iterator p;

    /*! iterate through all existing machines */
    for(p = MachineList.begin(); p != MachineList.end(); ++p)
    {
        CMachineNodeList *pCurNode = (*p)->GetNodeList();

        /*! apply movement caused by acceleration and gravity */
        while(pCurNode != 0)
        {
            float fPosX = pCurNode->GetPosX();
            float fPosY = pCurNode->GetPosY();

            float fPosXs = pCurNode->GetPosXs();
            float fPosYs = pCurNode->GetPosYs();

            //*
            if(fabs(fPosXs) < CLAMP_VELOCITY_MIN) 
            { 
                fPosXs = 0.0f; 
            }

            if(fabs(fPosYs) < CLAMP_VELOCITY_MIN) 
            { 
                fPosYs = 0.0f; 
            }

            if(fabs(fPosXs) > CLAMP_VELOCITY_MAX)
            {
                if(fPosXs > 0) { fPosXs = CLAMP_VELOCITY_MAX; } 
                else { fPosXs = -CLAMP_VELOCITY_MAX; }
            }

            if(fabs(fPosYs) > CLAMP_VELOCITY_MAX)
            {
                if(fPosYs > 0) { fPosYs = CLAMP_VELOCITY_MAX; } 
                else { fPosYs = -CLAMP_VELOCITY_MAX; }
            }
            //*/

            /*! @todo enforce gravity relative to arbitrary line */

            // center gravity
            /*
            if(fPosY > 0.5) { fPosYs -= fGravY * step; }
            else { fPosYs += fGravY * step; }

            if(fPosX > 0.5) { fPosXs -= fGravY * step; }
            else { fPosXs += fGravY * step; }
            //*/

            /*! apply gravity */
            fPosXs -= fGravX * step;
            fPosYs -= fGravY * step;

            /*! dampen velocity ("air resistance") */
            fPosXs *= AIR_RESISTANCE;
            fPosYs *= AIR_RESISTANCE;

            /*! clamp velocity */
            /*
            if(fPosXs > 0.2f) { fPosXs = 0.2f; }
            else if(fPosXs < -0.2f) { fPosXs = -0.2f; }

            if(fPosYs > 0.2f) { fPosYs = 0.2f; }
            else if(fPosYs < -0.2f) { fPosYs = -0.2f; }
            //*/

            /*! apply velocity */
            fPosX -= fPosXs;
            fPosY -= fPosYs;

            /*! clamp position ("bounce") */
            if(fPosX > 1.0f)
            {
                fPosX = 2.0f - fPosX;
                fPosXs = -fPosXs*0.75f;
            }
            else if(fPosX < 0.0f)
            {
                fPosX = 0.0f - fPosX;
                fPosXs = -fPosXs*0.75f;
            }

            if(fPosY > 1.0f)
            {
                fPosY = 2.0f - fPosY;
                fPosYs = -fPosYs*0.75f;
            }
            else if(fPosY < 0.0f)
            {
                fPosY = 0.0f - fPosY;
                fPosYs = -fPosYs*0.75f;
            }

            /*! apply position */
            pCurNode->SetPos(fPosX, fPosY);
            pCurNode->SetPoss(fPosXs, fPosYs);

            /*! advance to next node */
            pCurNode = pCurNode->pNext;
        }

        CMachineConnList *pCurConn = (*p)->GetConnList();

        /*! apply forces generated by connections for this machine */
        while(pCurConn != 0)
        {
            CMachineNode *pNode0 = pCurConn->GetNode0();
            CMachineNode *pNode1 = pCurConn->GetNode1();

            /*! retrieve node positions */
            float fPosX0 = pNode0->GetPosX();
            float fPosY0 = pNode0->GetPosY();
            float fPosX1 = pNode1->GetPosX();
            float fPosY1 = pNode1->GetPosY();

            /*! retrieve node accelerations */
            float fPosX0s = pNode0->GetPosXs();
            float fPosY0s = pNode0->GetPosYs();
            float fPosX1s = pNode1->GetPosXs();
            float fPosY1s = pNode1->GetPosYs();

            /*! calculate the delta in node positions */
            float diffX = fPosX1 - fPosX0;
            float diffY = fPosY1 - fPosY0;

            /*! calculate the distance between nodes */
            float distance = sqrtf(diffX*diffX + diffY*diffY);

            /*! @todo apply "muscles" */
            /*
            //if(pCurConn->pNext == NULL)
            {
                static float h = -3.1415f;
                tension += sin(h) / 100.0f;

                h += 0.1f;
            }
            */

            /*! calculate the amount of pull to apply based on tension */
            float pullDist = (pCurConn->GetLength() - distance) / TENSION; /*! @todo allow configurable tension (i.e. 20.0f) */

            float normDiffX = diffX / distance;
            float normDiffY = diffY / distance;

            fPosX0s += normDiffX * pullDist;
            fPosY0s += normDiffY * pullDist;

            fPosX1s -= normDiffX * pullDist;
            fPosY1s -= normDiffY * pullDist;

            pNode0->SetPoss(fPosX0s, fPosY0s);
            pNode1->SetPoss(fPosX1s, fPosY1s);

            /*! advance to next connection */
            pCurConn = pCurConn->pNext;
        }
    }
}
