// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#ifndef CRENDER_H
#define CRENDER_H

#include "BaseTypes.h"

/*! prototype used to update a single palette entry */
typedef void (*FadePaletteCallback)(void *pContext, uint32 i, float r, float g, float b);

/*! @brief CRender Palettes */
typedef enum _CRenderPalette
{
    CRENDER_PALETTE_RANDOM_SMOOTH_HUE = 0,
    CRENDER_PALETTE_BROKEN_TV,
    CRENDER_PALETTE_RANDOM_POLARIZED_HUE,
    CRENDER_PALETTE_RANDOM_POLARIZED_HUE2,
    CRENDER_PALETTE_COUNT,

    CRENDER_PALETTE_RANDOM = -1,
}
CRenderPalette;

/*! @brief Generic rendering base class
 *
 *  Renderers should inherit from this base class to provide the ability to
 *  render 8-bit graphics to the screen (or other device)
 *
 */

class CRender
{
    public:

        virtual ~CRender() { }

        /*! render pPixels using currently palette */
        virtual bool RenderPixels(uint08 *pPixels) = 0;

        /*! rotate palette one step */
        virtual bool RotatePalette() = 0;

        /*! set next palette */
        virtual void SetNextPalette(CRenderPalette nextPal) { this->nextPal = nextPal; }

        /*! get next palette */
        virtual CRenderPalette GetNextPalette() { return this->nextPal; }

        /*! enable stereogram rendering */
        virtual void SetStereogramEnabled(bool bEnable = true) { this->bStereogram = bEnable; }

        /*! check if stereogram is enabled */
        virtual bool GetStereogramEnabled() { return this->bStereogram; }

    protected:

        /*! convert from HSV colorspace to RGB colorspace */
        static void hsv2rgb(float fH, float fS, float fV, float *fR, float *fG, float *fB);

        /*! fade palette between two RGB triplets */
        static void FadePaletteFL(float r0, float g0, float b0, float r1, float g1, float b1, FadePaletteCallback pCallback, void *pContext, int first = 0, int last = 255);

        /*! fade palette between two RGB triplets */
        static void FadePaletteDW(uint32 r0, uint32 g0, uint32 b0, uint32 r1, uint32 g1, uint32 b1, FadePaletteCallback pCallback, void *pContext, int first = 0, int last = 255);

        /*! generate a palette */
        void GeneratePalette(FadePaletteCallback pCallback, void *pContext);

        /*! protected constructor to prevent base class instantiation */
        CRender(uint width, uint height);

        /*! cached rendering dimensions */
        uint width, height;

        /*! next palette */
        CRenderPalette nextPal;

        /*! flag that determines if we should render as a stereogram */
        bool bStereogram;
};

#endif
