// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "CRenderDD4.h"

CRenderDD4::CRenderDD4(uint width, uint height, HWND hWnd) : CRender(width, height), hWnd(hWnd)
{
    pDirectDraw = 0;
    pDirectDraw4 = 0;
    pDD4SPrimary = 0;
    pDD4SBackBuffer = 0;
    pDD4Clipper = 0;

    rotPaletteOffset = 0;

    pColorTable = new uint32[256];
    pColorTableBack = new uint32[256];
}

CRenderDD4::~CRenderDD4()
{
    if(pDD4Clipper != NULL)
    {
        pDD4Clipper->Release();
        pDD4Clipper = 0;
    }

    if(pDD4SBackBuffer != NULL)
    {
        pDD4SBackBuffer->Release();
        pDD4SBackBuffer = 0;
    }

    if(pDD4SPrimary != NULL)
    {
        pDD4SPrimary->Release();
        pDD4SPrimary = 0;
    }

    if(pDirectDraw4 != NULL)
    {
        pDirectDraw4->Release();
        pDirectDraw4 = 0;
    }

    if(pDirectDraw != NULL)
    {
        pDirectDraw->Release();
        pDirectDraw = 0;
    }

    if(pColorTable != 0)
    {
        delete[] pColorTable;
    }

    if(pColorTableBack != 0)
    {
        delete[] pColorTableBack;
    }
}

bool CRenderDD4::Create()
{
    HRESULT hRet = DirectDrawCreate(NULL, &this->pDirectDraw, NULL);

    if(FAILED(hRet))
        return false;

    hRet = this->pDirectDraw->QueryInterface(IID_IDirectDraw4, (PVOID*)&this->pDirectDraw4);

    if(FAILED(hRet))
        return false;

    hRet = this->pDirectDraw4->SetCooperativeLevel(hWnd, DDSCL_NORMAL);

    if(FAILED(hRet))
        return false;

    DDSURFACEDESC2 surfaceDesc;

    ZeroMemory(&surfaceDesc, sizeof(DDSURFACEDESC2));

    surfaceDesc.dwSize = sizeof(DDSURFACEDESC2);
    surfaceDesc.dwFlags = DDSD_CAPS;
    surfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

    hRet = pDirectDraw4->CreateSurface(&surfaceDesc, &this->pDD4SPrimary, NULL);

    if(FAILED(hRet))
        return false;

    hRet = pDirectDraw4->CreateClipper(0, &pDD4Clipper, NULL);

    if(FAILED(hRet))
        return false;

    hRet = pDD4Clipper->SetHWnd(0, hWnd);

    if(FAILED(hRet))
        return false;

    hRet = pDD4SPrimary->SetClipper(pDD4Clipper);

    if(FAILED(hRet))
        return false;

    ZeroMemory(&surfaceDesc, sizeof(DDSURFACEDESC2));

    surfaceDesc.dwSize = sizeof(DDSURFACEDESC2);
    surfaceDesc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
    surfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    surfaceDesc.dwWidth = width;
    surfaceDesc.dwHeight = height;

    hRet = pDirectDraw4->CreateSurface(&surfaceDesc, &this->pDD4SBackBuffer, NULL);

    if(FAILED(hRet))
        return false;

    /*! initialize random front palette */
    GeneratePalette(fpcRGB32, pColorTable);

    /*! initialize random back palette */
    GeneratePalette(fpcRGB32, pColorTableBack);

    return true;
}

bool CRenderDD4::RenderPixels(uint08 *pPixels)
{
    HRESULT hRet;

    DDSURFACEDESC2 surfaceDesc;

    ZeroMemory(&surfaceDesc, sizeof(DDSURFACEDESC2));

    surfaceDesc.dwSize = sizeof(DDSURFACEDESC2);

    if(pDD4SBackBuffer == NULL)
        return false;

    hRet = pDD4SBackBuffer->Lock(NULL, &surfaceDesc, DDLOCK_WAIT, NULL);

    if(FAILED(hRet))
    {
        if(hRet == DDERR_SURFACELOST) 
        {
            pDD4SBackBuffer->Restore();
        }

        return false;
    }

    uint32 *xrgb = (uint32*)surfaceDesc.lpSurface;
    uint32  wadd = (surfaceDesc.lPitch / 4) - width;

    int x, y, w = width, h = height;

    if(!bStereogram)
    {
        for(y=0;y<h;y++)
        {
            for(x=0;x<w;x++)
            {
                *xrgb++ = pColorTable[*pPixels++];
            }

            xrgb += wadd;
        }
    }
    else
    {
        static float dist = 2.0f;
        static float distS = 0.04f;

        if(dist > 5.0f || dist < 2.0f)
            distS = -distS;

        dist += distS;

        #define round(X) (int)((X)+0.5)
        #define DPI 72
        #define E round(dist*DPI)
        #define mu (1/3.0f)
        #define separation(Z) round((1-mu*Z)*E/(2-mu*Z))

        static uint32 *pix = 0;
        static int *same = 0;
    
        if(pix == 0)
        {
            pix = (uint32*)malloc(w*sizeof(uint32));
        }

        if(same == 0)
        {
            same = (int*)malloc(w*sizeof(int));
        }

        for(y=0;y<h;y++)
        {
            int s, left, right;

            for(x=0;x<w;x++)
            {
                same[x] = x;
            }

            for(x=0;x<w;x++)
            {
                float z = (float)(*pPixels) / 255.0f;

                s = separation(z);

                left = x - s/2;
                right = left + s;

                if(0 <= left && right < w)
                {
                    bool visible = true;

                    if(false)
                    {
                        int t=1;

                        float zt;

                        do
                        {
                            float zmt = (float)(*(pPixels - t)) / 255.0f;
                            float zpt = (float)(*(pPixels + t)) / 255.0f;

                            zt = z + 2*(2 - mu*z)*t/(mu*E);

                            visible = (zmt < zt) && (zpt < zt);

                            t++;
                        }
                        while(visible && zt < 1);
                    }

                    if(visible)
                    {
                        int remLeft = left;
                        int remRight = right;
                        int l = same[left];

                        while(l != left && l != right)
                        {
                            if(l < right)
                            {
                                left = l;
                                l = same[left];
                            }
                            else
                            {
                                same[left] = right;
                                left = right;
                                l = same[left];
                                right = l;
                            }
                        }

                        same[left] = right;
                    }
                }

                pPixels++;
            }

            for(x=0;x<w;x++)
            {
                if(same[x] == x)
                {
                    pix[x] = (rand()%2 == 0) ? 0x00000000 : 0x00FFFFFF;
                    //pix[x] = (rand()%2 == 0) ? 0x00000044 : 0x00FFFFFF;
                    //pix[x] = (rand() | (rand() << 16)) & 0x00FFFFFF;
                }
                else
                {
                    pix[x] = pix[same[x]];
                }

                *xrgb++ = pix[x];
            }

            xrgb += wadd;
        }
    }

    //free(pix);
    //free(same);

    pDD4SBackBuffer->Unlock(NULL);

    RECT destRect, srcRect, barRect1 = {0}, barRect2 = {0};

    /*! calculate source rectangle */
    SetRect(&srcRect, 0, 0, width, height);

    /*! calculate destination rectangle */
    {
        LONG offsX = 0, offsY = 0;

        POINT p = {0, 0};

        ClientToScreen(hWnd, &p);

        GetClientRect(hWnd, &destRect);

        double srcRatio = (double)width/(double)height;
        double dstRatio = (double)destRect.right/(double)destRect.bottom;

        if(srcRatio < dstRatio)
        {
            LONG newRight = (LONG)(destRect.bottom * srcRatio);

            offsX = (destRect.right - newRight) / 2;

            destRect.right = newRight;

            SetRect(&barRect1, 0, 0, offsX, destRect.bottom);
            SetRect(&barRect2, offsX + newRight, 0, offsX + offsX + newRight, destRect.bottom);
        }
        else if(srcRatio > dstRatio)
        {
            LONG newBottom = (LONG)(destRect.right / srcRatio);

            offsY = (destRect.bottom - newBottom) / 2;

            destRect.bottom = newBottom;

            SetRect(&barRect1, 0, 0, destRect.right, offsY);
            SetRect(&barRect2, 0, offsY + newBottom, destRect.right, offsY + offsY + newBottom);
        }

        OffsetRect(&destRect, p.x + offsX, p.y + offsY);

        OffsetRect(&barRect1, p.x, p.y);
        OffsetRect(&barRect2, p.x, p.y);
    }

    hRet = pDD4SPrimary->Blt(&destRect, pDD4SBackBuffer, &srcRect, DDBLT_ASYNC, NULL);

    /*! restore lost surface, if necessary */
    if(hRet == DDERR_SURFACELOST) 
    {
        pDD4SBackBuffer->Restore();
        pDD4SPrimary->Restore(); 
    }

    /*! draw black bars, if necessary */
    {
        DDBLTFX ddbltfx;

        ZeroMemory(&ddbltfx, sizeof(DDBLTFX));

        ddbltfx.dwSize = sizeof(DDBLTFX);
        ddbltfx.dwFillColor = 0;

        if( (barRect1.left != barRect1.right) && (barRect1.top != barRect1.bottom) )
        {
            hRet = pDD4SPrimary->Blt(&barRect1, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &ddbltfx);
        }

        if( (barRect2.left != barRect2.right) && (barRect2.top != barRect2.bottom) )
        {
            hRet = pDD4SPrimary->Blt(&barRect2, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &ddbltfx);
        }
    }

    return true;
}

bool CRenderDD4::RotatePalette()
{
    if(rotPaletteOffset < 256*2)
    {
        //
        // NOTE: Solar Rotation
        //

        //pColorTable[rotPaletteOffset] = pColorTableBack[rotPaletteOffset];
        if(rotPaletteOffset%2 == 0)
        {
            for(int r=0;r<256;r++)
            {
                uint32 rs = pColorTableBack[r] & 0x00FF0000;
                uint32 gs = pColorTableBack[r] & 0x0000FF00;
                uint32 bs = pColorTableBack[r] & 0x000000FF;

                uint32 rd = pColorTable[r] & 0x00FF0000;
                uint32 gd = pColorTable[r] & 0x0000FF00;
                uint32 bd = pColorTable[r] & 0x000000FF;

                if(rd > rs) { rd -= 0x00010000; }
                else if(rd < rs) { rd += 0x00010000; }

                if(gd > gs) { gd -= 0x00000100; }
                else if(gd < gs) { gd += 0x00000100; }

                if(bd > bs) { bd -= 0x00000001; }
                else if(bd < bs) { bd += 0x00000001; }

                pColorTable[r] = rd | gd | bd;
            }
        }
    }
    else if(rotPaletteOffset == 256*2)
    {
        GeneratePalette(fpcRGB32, pColorTableBack);

        rotPaletteOffset = -1;
    }

    rotPaletteOffset++;

    return true;
}

void CRenderDD4::fpcRGB32(void *pContext, uint32 i, float r, float g, float b)
{
    uint32 *pColorTable = (uint32*)pContext;

    /*
    static bool    bCached = false;
    static uint08 *pRGB = NULL;

    if(!bCached)
    {
        FILE *kthx = fopen("C:\\palette.raw", "rb");

        pRGB = (uint08*)malloc(3*sizeof(uint08)*256);

        fread(pRGB, 3*sizeof(uint08)*256, 1, kthx);

        fclose(kthx);
    }
    //*/

    uint08 uiR = (uint08)(r*255.0f);
    uint08 uiG = (uint08)(g*255.0f);
    uint08 uiB = (uint08)(b*255.0f);

    /*
    uiR = pRGB[i*3+0];
    uiG = pRGB[i*3+1];
    uiB = pRGB[i*3+2];
    //*/

    pColorTable[i] = 0xFF000000 | (uiR << 16) | (uiG << 8) | uiB;

    return;
}
