// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "CRenderDX9Texture.h"

#include <math.h>

CRenderDX9Texture::CRenderDX9Texture(uint width, uint height, HWND hWnd) : CRender(width, height), hWnd(hWnd)
{
    pDirect3D9 = 0;
    pDirect3DDevice9 = 0;
    rotPaletteOffset = 0;

    pColorTable = new uint32[256];
    pColorTableBack = new uint32[256];
}

CRenderDX9Texture::~CRenderDX9Texture()
{
    if(pDirect3DTexture9 != NULL)
    {
        pDirect3DTexture9->Release();
        pDirect3DTexture9 = NULL;
    }

    if(pDirect3DDevice9 != NULL)
    {
        pDirect3DDevice9->Release();
        pDirect3DDevice9 = NULL;
    }

    if(pDirect3D9 != NULL)
    {
        pDirect3D9->Release();
        pDirect3D9 = NULL;
    }

    if(pColorTable != 0)
    {
        delete[] pColorTable;
        pColorTable = 0;
    }

    if(pColorTableBack != 0)
    {
        delete[] pColorTableBack;
        pColorTableBack = 0;
    }
}

bool CRenderDX9Texture::Create()
{
    pDirect3D9 = Direct3DCreate9(D3D_SDK_VERSION);

    if(pDirect3D9 == NULL)
    {
        return false;
    }

    D3DPRESENT_PARAMETERS PresentationParameters;

    ZeroMemory(&PresentationParameters, sizeof(PresentationParameters));

    PresentationParameters.BackBufferWidth = width;
    PresentationParameters.BackBufferHeight = height;
    PresentationParameters.BackBufferFormat = D3DFMT_UNKNOWN;
    PresentationParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
    PresentationParameters.hDeviceWindow = hWnd;
    PresentationParameters.Windowed = TRUE;
    PresentationParameters.EnableAutoDepthStencil = TRUE;
    PresentationParameters.AutoDepthStencilFormat = D3DFMT_D16;
    PresentationParameters.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
    PresentationParameters.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    PresentationParameters.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    HRESULT hRet = pDirect3D9->CreateDevice
    (
        D3DADAPTER_DEFAULT,
        D3DDEVTYPE_HAL,
        hWnd,
        D3DCREATE_HARDWARE_VERTEXPROCESSING, // @todo: Configure this..
        &PresentationParameters,
        &pDirect3DDevice9
    );

    if(FAILED(hRet))
    {
        return false;
    }

    pDirect3DDevice9->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    pDirect3DDevice9->SetRenderState(D3DRS_LIGHTING, FALSE);
    pDirect3DDevice9->SetRenderState(D3DRS_ZENABLE, TRUE);

    hRet = pDirect3DDevice9->CreateTexture(width, height, 1, 0, D3DFMT_X8R8G8B8, D3DPOOL_MANAGED, &pDirect3DTexture9, NULL);

    if(FAILED(hRet))
    {
        return false;
    }

    //
    // initialize palette front/back
    //

    GeneratePalette(fpcRGB32, pColorTable);
    GeneratePalette(fpcRGB32, pColorTableBack);

    return true;
}

bool CRenderDX9Texture::RenderPixels(uint08 *pPixels)
{
    pDirect3DDevice9->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);

    pDirect3DDevice9->BeginScene();

    /*! draw primitive */
    {
        float z = 0.0f;
        float rhw = 1.0f/(z*990.0f+10.0f);

        DWORD col = RGB(0xFF, 0xFF, 0xFF);

        SimpleVertex verts[4] =
        {
            0-0.5f,  0-0.5f,    z, rhw, col, 0.0f, 0.0f,
            width-0.5f, 0-0.5f,    z, rhw, col, 1.0f, 0.0f,
            width-0.5f, height-0.5f, z, rhw, col, 1.0f, 1.0f, 
            0-0.5f, height-0.5f, z, rhw, col, 0.0f, 1.0f
        };

        pDirect3DDevice9->SetTexture(0, pDirect3DTexture9);

        pDirect3DDevice9->SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);

        pDirect3DDevice9->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, verts, sizeof(SimpleVertex) );
    }

    pDirect3DDevice9->EndScene();

    D3DLOCKED_RECT LockedRect;

    HRESULT hRet = pDirect3DTexture9->LockRect(0, &LockedRect, NULL, 0);

    if(FAILED(hRet))
    {
        return false;
    }

    uint32 *xrgb = (uint32*)LockedRect.pBits;
    uint32  wadd = (LockedRect.Pitch / 4) - width;

    uint32 w = width, h = height;

    for(uint y=0;y<h;y++)
    {
        for(uint x=0;x<w;x++)
        {

            *xrgb++ = pColorTable[*pPixels++];
        }

        xrgb += wadd;
    }

    pDirect3DTexture9->UnlockRect(0);

    pDirect3DDevice9->Present(NULL, NULL, NULL, NULL);

    return true;
}

bool CRenderDX9Texture::RotatePalette()
{
    if(rotPaletteOffset < 256)
    {
        //
        // NOTE: Solar Rotation
        //

        pColorTable[rotPaletteOffset] = pColorTableBack[rotPaletteOffset];
    }
    else if(rotPaletteOffset == 256*5)
    {
        GeneratePalette(fpcRGB32, pColorTableBack);

        rotPaletteOffset = -1;
    }

    rotPaletteOffset++;

    return true;
}

void CRenderDX9Texture::fpcRGB32(void *pContext, uint32 i, float r, float g, float b)
{
    uint32 *pColorTable = (uint32*)pContext;

    uint08 uiR = (uint08)(r*255.0f);
    uint08 uiG = (uint08)(g*255.0f);
    uint08 uiB = (uint08)(b*255.0f);

    pColorTable[i] = 0xFF000000 | (uiR << 16) | (uiG << 8) | uiB;

    return;
}
