// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "CRenderXbox.h"

#include <xgraphics.h>

CRenderXbox::CRenderXbox(uint width, uint height) : CRender(width, height)
{
    pDirect3D8 = 0;
    pDirect3DDevice8 = 0;
    rotPaletteOffset = 256;

    pColorTable = new uint32[256];
    pColorTableBack = new uint32[256];
}

CRenderXbox::~CRenderXbox()
{
    if(pDirect3DDevice8 != NULL)
    {
        pDirect3DDevice8->Release();
        pDirect3DDevice8 = NULL;
    }

    if(pDirect3D8 != NULL)
    {
        pDirect3D8->Release();
        pDirect3D8 = NULL;
    }

    if(pColorTable != 0)
    {
        delete[] pColorTable;
        pColorTable = 0;
    }

    if(pColorTableBack != 0)
    {
        delete[] pColorTableBack;
        pColorTableBack = 0;
    }
}

bool CRenderXbox::Create()
{
    pDirect3D8 = Direct3DCreate8(D3D_SDK_VERSION);

    if(pDirect3D8 == NULL)
    {
        return false;
    }

    D3DPRESENT_PARAMETERS PresentationParameters;

    ZeroMemory(&PresentationParameters, sizeof(PresentationParameters));

    PresentationParameters.BackBufferWidth = width;
    PresentationParameters.BackBufferHeight = height;
    PresentationParameters.BackBufferFormat = D3DFMT_X8R8G8B8;
    PresentationParameters.BackBufferCount = 1;
    PresentationParameters.MultiSampleType = D3DMULTISAMPLE_NONE;
    PresentationParameters.SwapEffect = D3DSWAPEFFECT_FLIP;
    PresentationParameters.EnableAutoDepthStencil = FALSE;
    PresentationParameters.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
    PresentationParameters.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;

    HRESULT hRet = pDirect3D8->CreateDevice
    (
        D3DADAPTER_DEFAULT,
        D3DDEVTYPE_HAL,
        NULL,
        D3DCREATE_HARDWARE_VERTEXPROCESSING,
        &PresentationParameters,
        &pDirect3DDevice8
    );

    if(FAILED(hRet))
    {
        return false;
    }

    //
    // initialize palette front/back
    //

    GeneratePalette(fpcRGB32, pColorTable);
    GeneratePalette(fpcRGB32, pColorTableBack);

    return true;
}

bool CRenderXbox::RenderPixels(uint08 *pPixels)
{
    IDirect3DSurface8 *pSurface8 = NULL;

    HRESULT hRet = pDirect3DDevice8->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &pSurface8);

    if(FAILED(hRet))
    {
        return false;
    }

    D3DLOCKED_RECT LockedRect;

    hRet = pSurface8->LockRect(&LockedRect, NULL, D3DLOCK_TILED);

    if(FAILED(hRet))
    {
        pSurface8->Release();
        return false;
    }


    uint32 *xrgb = (uint32*)LockedRect.pBits;

    uint32 wadd = (LockedRect.Pitch / 4) - width;
    int x, y, w = width, h = height;

    for(y=0;y<h;y++)
    {
        for(x=0;x<w;x++)
        {
            *xrgb++ = pColorTable[*pPixels++];
        }

        xrgb += wadd;
    }

    pSurface8->UnlockRect();

    pSurface8->Release();

    pDirect3DDevice8->Present(NULL, NULL, NULL, NULL);

    return true;
}

bool CRenderXbox::RotatePalette()
{
    if(rotPaletteOffset < 256*5)
    {
        //
        // NOTE: Solar Rotation
        //

        //pColorTable[rotPaletteOffset] = pColorTableBack[rotPaletteOffset];
        if(rotPaletteOffset%5 == 0)
        {
            for(int r=0;r<256;r++)
            {
                uint32 rs = pColorTableBack[r] & 0x00FF0000;
                uint32 gs = pColorTableBack[r] & 0x0000FF00;
                uint32 bs = pColorTableBack[r] & 0x000000FF;

                uint32 rd = pColorTable[r] & 0x00FF0000;
                uint32 gd = pColorTable[r] & 0x0000FF00;
                uint32 bd = pColorTable[r] & 0x000000FF;

                if(rd > rs) { rd -= 0x00010000; }
                else if(rd < rs) { rd += 0x00010000; }

                if(gd > gs) { gd -= 0x00000100; }
                else if(gd < gs) { gd += 0x00000100; }

                if(bd > bs) { bd -= 0x00000001; }
                else if(bd < bs) { bd += 0x00000001; }

                pColorTable[r] = rd | gd | bd;
            }
        }
    }
    else if(rotPaletteOffset == 256*6)
    {
        GeneratePalette(fpcRGB32, pColorTableBack);

        rotPaletteOffset = -1;
    }

    rotPaletteOffset++;

    return true;
}

void CRenderXbox::fpcRGB32(void *pContext, uint32 i, float r, float g, float b)
{
    uint32 *pColorTable = (uint32*)pContext;

    uint08 uiR = (uint08)(r*255.0f);
    uint08 uiG = (uint08)(g*255.0f);
    uint08 uiB = (uint08)(b*255.0f);

    pColorTable[i] = 0xFF000000 | (uiR << 16) | (uiG << 8) | uiB;

    return;
}
