// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// *****************************************************************

#ifndef AUTOMUTEXRW_H
#define AUTOMUTEXRW_H

#include "MutexRW.h"

/*! @brief Automatic mutex manager for multiple readers and a single writer
 * 
 * Automatic helper object that manages the locking and unlocking of a mutex
 * when the heap allocated instance goes out of scope. In the following example,
 * the mutex will lock at "// A" and unlock at "// B" :
 *
 * <pre>
 * ...
 * {
 *     AutoMutexRW lock(&mutex, true); // A
 *     ...
 * } // B
 * ...
 * </pre>
 *
 */

class AutoMutexRW
{
    public:

        /*! Constructor, requires an initialized mutex instance */
        AutoMutexRW(MutexRW *pMutexRW, bool writer = false, bool enable = true) : pMutexRW(pMutexRW), writer(writer), done(!enable)
        {
            /*! optionally, you can disable use of this mutex */
            if(!enable) { return; }

            if(pMutexRW != NULL)
            {
                if(writer) { pMutexRW->WaitW(); }
                else { pMutexRW->WaitR(); }
            }
        }

        /*! Deconstructor */
       ~AutoMutexRW() { Release(); }

        /*! Prevents double release */
        void EarlyRelease() { Release(); }

    private:

        /*! Release MutexRW */
        void Release()
        {
            if(done) { return; }

            if(pMutexRW != NULL)
            {
                if(writer) { pMutexRW->ReleaseW(); }
                else { pMutexRW->ReleaseR(); }
            }

            done = true;
        }

        /*! local cache of mutex instance */
        MutexRW *pMutexRW;

        /*! local cache of read/write configuration */
        bool writer;

        /*! have we released? */
        bool done;
};

#endif
