// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#ifndef THREAD_H
#define THREAD_H

#include "BaseTypes.h"

/*! generic thread function */
typedef void (*ThreadFunc)(bool *pStopFlag, void *pContext);

/*! @brief Simple thread object
 *
 *  This thread object provides a platform independant way to
 *  Start/Stop a thread of execution. To use this thread object,
 *  the provided ThreadFunc must periodically check the stop
 *  flag to know when to stop executing (return).
 *
 */

class Thread
{
    public:

        Thread(ThreadFunc pfFunc, void *pContext);
       ~Thread();

        /*! Start thread */
        void Start();

        /*! Stop thread */
        void Stop(bool async = false);

        /*! Access thread function */
        ThreadFunc GetFunc() { return pfFunc; }

        /*! Access thread context */
        void *GetThreadContext() { return pContext; }

        /*! Access the stop flag */
        bool GetStopFlag() { return bStop; }

        /*! Access the address of stop flag */
        bool *GetStopFlagRef() { return &bStop; }

        /*! Access stop event */
        void *GetStoppedEvent() { return pStoppedEvent; }

        /*! Is this thread stopped? */
        bool IsStopped();

        /*! Sleep the current thread for specified number of milliseconds */
        static void Sleep(int ms);

        /*! Retrieve unique thread identifier */
        static int GetCurrentThreadID();

    private:

        /*! Cached thread function */
        ThreadFunc pfFunc;

        /*! Cached arbitrary user context */
        void *pContext;

        /*! Cached thread handle */
        void *pHandle;

        /*! used to signal stopping the thread */
        bool bStop;

        /*! event triggered when thread is dead */
        void *pStoppedEvent;
};

/*! Utility macro for 'this' call function forwarding */
#define THREAD_FUNC_FWD(ClassName, FuncNameFwd, FuncName)           \
static void FuncNameFwd(bool *pStopFlag, void *pContext)            \
{                                                                   \
    ClassName *pThis = (ClassName*)pContext;                        \
                                                                    \
    if(pThis == NULL) { return; }                                   \
                                                                    \
    pThis->FuncName(pStopFlag);                                     \
    return;                                                         \
}

#endif

