// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#ifndef SIRENIC_H
#define SIRENIC_H

#include "Assign.h"
#include "RenderFX.h"
#include "Transform.h"
#include "TransformConfig.h"

#include "Mutex.h"
#include "Thread.h"

#define _SIRENIC_VERSION "0.5.0"

#include <math.h>

/*! @mainpage Sirenic Visualization Library
 *
 *  <center>
 *  \image html megadance.gif
 *  </center>
 *  <br>
 *  Navigate above to view documentation.
 *  <br>
 *  <br>
 */

/*! @brief Sirenic core system 
 *
 *  Encapsulates most of the functionality of the Sirenic project. This
 *  class allows for the processing mouse state, rendering various effects,
 *  drawing various lines/points/curves/etc, and interpolated transformations
 *
 */

class Sirenic
{
    public:

        /*! construct a sirenic core with the specified dimensions and buffers */
        Sirenic(int width, int height, int pitch, int bufferCount = 2);

        virtual ~Sirenic();

        /*! initialize the system */
        bool Create();

        /*! \name Mouse interface */
        /*! \{ */

        /*! set mouse position */
        void SetMousePos(uint mouseX, uint mouseY) { this->mouseX = mouseX; this->mouseY = mouseY; }

        /*! get mouse X position */
        uint GetMousePosX() { return this->mouseX; }

        /*! get mouse Y position */
        uint GetMousePosY() { return this->mouseY; }

        /*! set mouse left button down state */
        void SetMouseDownL(bool mouseDownL) { this->mouseDownL = mouseDownL; }

        /*! get mouse left button down state */
        bool GetMouseDownL() { return this->mouseDownL; }

        /*! \} */

        /*! \name Render FX interface */
        /*! \{ */

        /*! insert/remove render effect */
        RenderFXNode *InsertRenderFX(RenderFX *pRenderFX);
        bool RemoveRenderFX(RenderFXNode *pRenderFXNode);

        /*! draw all queued render effects */
        bool DrawRenderFX();

        /*! \} */

        /*! \name Drawing interface */
        /*! \{ */

        /*! draw a single point */
        bool DrawPoint(int x0, int y0, uint08 c);

        /*! draw a line */
        bool DrawLine(int x0, int y0, int x1, int y1, uint08 c);

        /*! draw an antialiased line */
        bool DrawLineWu(float x0, float y0, float x1, float y1, float c);

        /*! draw a bezier curve */
        bool DrawBezier(CartCoord fCoord[4], float fStep, uint08 c);

        /*! draw a rectangle */
        bool DrawRect(int x0, int y0, int x1, int y1, uint08 c);

        /*! draw an antialiased rectangle */
        bool DrawRectWu(float x0, float y0, float x1, float y1, float c);

        /*! draw a filled rectangle */
        bool FillRect(int x0, int y0, int x1, int y1, uint08 c);

#ifndef _XBOX /*! @todo support DrawString on Xbox */
        /*! draw a string of text */
        bool DrawString(int x0, int y0, const char *str, int size, bool antiAlias, uint08 c);
#endif

        /*! Set assignment function */
        void SetAssignFunc(roAssignFunc assignFunc) { this->assignFunc = assignFunc; }

        /*! Access assignment function */
        roAssignFunc GetAssignFunc() { return this->assignFunc; }

        /*! \} */

        /*! \name Waveform interface */
        /*! \{ */

        /*! enable automatic waveform rendering */
        void EnableWaveform() { this->bWaveformEnabled = true; }

        /*! disable automatic waveform rendering */
        void DisableWaveform() { this->bWaveformEnabled = false; }

        /*! Access waveform enabled status */
        bool isWaveformEnabled() { return this->bWaveformEnabled; }

        /*! update waveform rendering data */
        void UpdateWaveform(uint08 *pNewWaveform);

        /*! update waveform rendering scale */
        void UpdateWaveformScale(float scale);

        /*! update waveform rendering color */
        void UpdateWaveformColor(float c) { this->waveformColor = c; }

        /*! \} */

        /*! \name Transform interface */
        /*! \{ */

        /*! set extra source weight flag specifying if extra source weight should be used*/
        void SetExtraSrcWeightFlag(bool flag) { this->bExtraSrcWeight = flag; }

        /*! retrieve flag specifying if extra source weight should be used */
        bool GetExtraSrcWeightFlag() { return this->bExtraSrcWeight; }

        /*! perform interpolated transformation */
        bool Transform();

        /*! \} */

        /*! \name Buffer interface */
        /*! \{ */

        /*! lock view buffer (returns raw pixel data) */
        uint08 *LockViewBuffer();

        /*! unlock view buffer */
        void UnlockViewBuffer();

        /*! retrieve view buffer, without locking (can have negative side effects) */
        uint08 *GetViewBuffer();

        /*! advance the current view buffer */
        void AdvanceViewBuffer();

        /*! \} */

        /*! Access width dimension */
        int GetWidth() { return this->width; }

        /*! Access height dimension */
        int GetHeight() { return this->height; }

        /*! Access pitch dimension */
        uint GetPitch() { return this->pitch; }

        /*! Access creation of system status */
        bool isCreated() { return this->bCreated; }

    private:

        /*! background thread for creating distortion fields */
        THREAD_FUNC_FWD(Sirenic, DistortThreadFwd, DistortThread);

        /*! linear interpolation in cartesian coordinates */
        static void LinearInterpolate(float fT, CartCoord &fCoord0, CartCoord &fCoord1, CartCoord &fOut);

        /*! forwarded from DistortThreadFwd to inherit 'this' pointer */
        void DistortThread(bool *pStopFlag);

        /*! fast line renderer */
        static void roDrawLine(uint08 *pBuffer, uint uiPitch, uint x0, uint y0, uint x1, uint y1, uint08 c, roAssignFunc assignFunc);

        /*! fast antialiased line renderer */
        static void roDrawLineWu(uint08 *pBuffer, uint uiPitch, float x0, float y0, float x1, float y1, float c);

        /*! fast rectangle renderer */
        static void roFillRect(uint08 *pBuffer, uint uiPitch, uint x0, uint y0, uint x1, uint y1, uint08 c);

        /*! fast pixel renderer */
        static inline void roDrawPixel(uint08 *pBuffer, uint uiPitch, uint x0, uint y0, uint08 c)
        {
            pBuffer[y0*uiPitch + x0] = c;
        }

        /*! fast antialiased pixel renderer */
        static inline void roDrawPixelWu(uint08 *pBuffer, uint uiPitch, float x0, float y0, float c)
        {
            float fx = x0 - (float)(int)x0;
            float fy = y0 - (float)(int)y0;

            float btl = (1 - fx) * (1 - fy) * c * 255.0f;
            float btr = (    fx) * (1 - fy) * c * 255.0f;
            float bbl = (1 - fx) * (    fy) * c * 255.0f;
            float bbr = (    fx) * (    fy) * c * 255.0f;

            pBuffer[((int)y0  )*uiPitch + (int)x0    ] = (uint08)btl;
            pBuffer[((int)y0  )*uiPitch + (int)x0 + 1] = (uint08)btr;
            pBuffer[((int)y0+1)*uiPitch + (int)x0    ] = (uint08)bbl;
            pBuffer[((int)y0+1)*uiPitch + (int)x0 + 1] = (uint08)bbr;
        }

#ifndef _XBOX /*! @todo support DrawString on Xbox */
        /*! fast text string renderer */
        static void roDrawString(uint08 *pBuffer, uint uiPitch, uint uiHeight, uint x0, uint y0, const char *str, int size, bool antiAlias, uint08 c);
#endif
        /*! create delta field */
        bool Distort(bool *pStopFlag, struct Delta *pDeltaField);

        /*! handle to distortion thread */
        Thread *pDistortThread;

        /*! current efault assignment function */
        roAssignFunc assignFunc;

        /*! current delta field generator */
        TransformConfig *transConfig;

        /*! doubly linked list of render effects */
        RenderFXNode *pRenderFXNodeH;

        /*! cached dimension information */
        int width, height, pitch;

        /*! cached number of buffers used by the system */
        int bufferCount;

        /*! cached mouse x,y postition */
        uint mouseX, mouseY;

        /*! cached status of left mouse button */
        bool mouseDownL;

        /*! array of pixel buffers */
        struct LockableBuffer *pBuffer;

        /*! current front delta field */
        struct Delta *pDeltaFieldFront;

        /*! current back delta field */
        struct Delta *pDeltaFieldBack;

        /*! dither map */
        sint08 *pDitherMap;

        /*! offset of view buffer into the pBuffer array */
        int viewBuffer;

        /*! offset of active buffer into the pBuffer array */
        int activeBuffer;

        /*! offset of previous buffer into the pBuffer array */
        int previousBuffer;

        /*! is waveform currently enabled? */
        bool bWaveformEnabled;
        /*! raw waveform data (two buffers) */
        uint08 *pWaveform[2];
        /*! current active waveform buffer */
        uint activeWaveform;
        /*! current waveform scale */
        float waveformScale;
        /*! current waveform color */
        float waveformColor;

        /*! should we transform with extra source weight? */
        bool bExtraSrcWeight;

        /*! protects pixel data buffers */
        Mutex *pBufferMutex;
        /*! protects render effects */
        Mutex *pRenderFXMutex;
        /*! protects waveform buffers */
        Mutex *pWaveformMutex;
        /*! locks view buffer */
        Mutex *pViewMutex;

        /*! cached creation status of this system */
        bool bCreated;

        /*! should transformation function darken the center? */
        bool darkenCenter;

        /*! should transformation function darken the edges? */
        bool darkenEdges;
};

/*! \brief Represents a single delta field entry */
struct Delta
{
    /*! x,y precombined offset */
    uint32 o;

    //
    // weight values
    //
    //     [w0]
    // [w1][  ][w2]
    //     [w3]
    //

    /*! weight of y-1 pixel */
    uint08 w0;
    /*! weight of x-1 pixel */
    uint08 w1;
    /*! weight of x+1 pixel */
    uint08 w2;
    /*! weight of y+1 pixel */
    uint08 w3;
};

/*! \brief Buffer with protection mutex */
struct LockableBuffer
{
    /*! mutex used to lock this buffer */
    Mutex  *pMutex;

    /*! raw buffer */
    uint08 *pPixels;
};

#endif
