// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#ifndef TRANSFORMCONFIG_H
#define TRANSFORMCONFIG_H

#include "BaseTypes.h"

#include <math.h>
#include <time.h>
#include <stdlib.h>

#define MAX_TRANSFORM_CONFIG_BYTES (64)

/*! @brief Transformation configuration
 *
 *  Abstract class for implementing a custom transformation configuration. Objects
 *  can inherit from this to provide a per-pixel transformation function.
 *
 */

class TransformConfig
{
    public:

        TransformConfig() { srand(time(NULL)); }

        /*!
         *  Transform function
         *
         *  @param pullX strength of pull in the X direction
         *  @param pullY strength of pull in the Y direction
         *  @param fX current X position
         *  @param fY current Y position
         *  @param pullR strength of pull in the R direction (polar coordinates)
         *  @param pullD strength of pull in the D direction (polar coordinates)
         *  @param fR current R position (polar coordinates)
         *  @param fD current D position (polar coordinates)
         *
         */

        virtual void Transform
        (
            float &pullX, float &pullY, float fX, float fY,
            float &pullR, float &pullD, float fR, float fD
        ) = 0;

        /*! Generate a random number between min and max */
        static inline float RandomRange(float min, float max)
        {
            return ((float)rand() / RAND_MAX) * (max - min) + min;
        }

        /*! Generate a random boolean */
        static inline bool RandomBool()
        {
            return (rand()%2 == 1);
        }

    protected:

        /*! \name Polar Coordinate Transformations */
        /*! \{ */

        static inline void dm_PolarPull(float &pullR, float fR, float pullStrength)
        {
            pullR += fR*pullStrength - fR;
        }

        static inline void dm_PolarWaves(float &pullR, float &pullD, float fR, float rippleCount, float rippleStrength)
        {
            pullD += (1-sinf(fR*rippleCount*3.1415f))*(1-fR)*rippleStrength;
            pullD -= 0.0025f;
        }

        static inline void dm_TwirlBump(float &pullR, float &pullD, float fR, float bumpCount, float bumpStrength, float twirlStrength)
        {
            pullR += sinf(fR*bumpCount*3.1415f)/((125/bumpStrength)*bumpCount);

            if(twirlStrength != 0)
            {
                pullD += sinf(1-fR)/(sinf(fR)*(125/twirlStrength));
            }
        }

        /*
        static inline void dm_AlienEggs(float &pullR, float &pullD, float fX, float fY, float countX, float countY, float strengthX, float strengthY)
        {
            pullR += sin(fX*countX*3.1415)/(1250/strengthX);
            pullR += cos(fY*countY*3.1415)/(1250/strengthY);
        }
        //*/

        /*! \} */

        /*! \name Rectangular Coordinate Transformations */
        /*! \{ */

        static inline void dm_RotoZoom(float &pullX, float &pullY, float fX, float fY, float strengthX, float strengthY)
        {
            pullX += (-fX)*strengthX - (-fY)*strengthX;
            pullY += (-fX)*strengthY + (-fY)*strengthY;
        }

        /*
        static inline void dm_Pull(float &pullX, float &pullY, float fX, float fY, float strengthX, float strengthY)
        {
            pullX += fX*strengthX;
            pullY += fY*strengthY;
        }

        static inline void dm_SlicerAdd(float &pullX, float &pullY, float fX, float fY, float countX, float countY, float strengthX, float strengthY)
        {
            if(fmod(fY+1, 4/countX) < (2/countX))
            {
                pullX -= strengthX;
            }
            else
            {
                pullX += strengthX;
            }

            if(fmod(fX+1, 4/countY) < (2/countY))
            {
                pullY -= strengthY;
            }
            else
            {
                pullY += strengthY;
            }
        }

        static inline void dm_Fuzz(float &pullX, float &pullY, float fX, float fY, float strengthX, float strengthY)
        {
            pullX += ((float)rand() / RAND_MAX)*strengthX - (strengthX/2);
            pullY += ((float)rand() / RAND_MAX)*strengthY - (strengthY/2);
        }

        static inline void dm_Bisine(float &pullX, float &pullY, float fX, float fY, float countX, float countY, float strengthX, float strengthY)
        {
            pullX += ( sin(fY*countX*3.1415) + sin(fX*countX*3.1415) )*strengthX;
            pullY += ( sin(fX*countY*3.1415) - sin(fY*countY*3.1415) )*strengthY;
        }

        static inline void dm_CircleRotation(float &pullX, float &pullY, float fX, float fY, float circleCountX, float circleCountY, float rotStrengthX, float rotStrengthY, bool bClockwise = true)
        {
            float circX = fmod(fX*circleCountX, 2) + (fX > 0 ? -1.0 : 1.0);
            float circY = fmod(fY*circleCountY, 2) + (fY > 0 ? -1.0 : 1.0);

            pullX += (bClockwise ?  circY : -circY)*rotStrengthX;
            pullY += (bClockwise ? -circX :  circX)*rotStrengthY;
        }

        static inline void dm_SquarePoppers(float &pullX, float &pullY, float fX, float fY, float squareCountX, float squareCountY, float strengthX, float strengthY)
        {
            pullX += (fmod(fX, 2.0 / squareCountX) - ((fX > 0) ? (1 / squareCountX) : -(1 / squareCountX)))*strengthX;
            pullY += (fmod(fY, 2.0 / squareCountY) - ((fY > 0) ? (1 / squareCountY) : -(1 / squareCountY)))*strengthY;
        }
        //*/
};

/*! @brief "Simple" transformation configuration */
class tc_Simple : public TransformConfig
{
    public:

        tc_Simple();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);
};

/*! @brief "Polar Waves" transformation configuration */
class tc_PolarWaves : public TransformConfig
{
    public:

        tc_PolarWaves();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

        float pullStrength;
        float rippleStrength;
        float rippleCount;
};

/*! @brief "Bumpy Vortex" transformation configuration */
class tc_BumpyVortex : public TransformConfig
{
    public:

        tc_BumpyVortex();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

        float pullStrength;
        float bumpStrength;
        float bumpCount;
        float twistStrength;
};

/*! @brief "Roto-Zoom" transformation configuration */
class tc_RotoZoom : public TransformConfig
{
    public:

        tc_RotoZoom();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

        float pullStrengthX;
        float pullStrengthY;
        float adjX;
        float adjY;
};

/*! @brief "Corner Stretch" transformation configuration */
class tc_CornerStretch : public TransformConfig
{
    public:

        tc_CornerStretch();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

        float pullStrength;
};

/*! @brief "Quadrant Rotate" transformation configuration */
class tc_QuadrantRotate : public TransformConfig
{
    public:

        tc_QuadrantRotate();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

        float pullStrengthQuad1;
        float pullStrengthQuad2;
        float pullStrengthQuad3;
        float pullStrengthQuad4;
};

/*! @brief "UnNamed" transformation configuration */
class tc_UnNamed : public TransformConfig
{
    public:

        tc_UnNamed();

        virtual void Transform(float &pullX, float &pullY, float fX, float fY, float &pullR, float &pullD, float fR, float fD);

    private:

};


            /* OLD SCHOOL SNIPPET */
            /*
            switch(rand()%12)
            {
                case 0:
                    TransformConfig::dm_SquarePoppers(pullX, pullY, fX, fY, 15, 15, 0.02f, 0.02f);
                    break;
                case 1:
                    TransformConfig::dm_Bisine(pullX, pullY, fX, fY, 8, 8, 0.002f, 0.002f);
                    break;
                case 2:
                    TransformConfig::dm_Pull(pullX, pullY, fX, fY, -0.02f, -0.005f);
                    break;
                case 3:
                    TransformConfig::dm_CircleRotation(pullX, pullY, fX, fY, 10, 10, 0.002f, 0.002f, true);
                    break;
                case 4:
                    TransformConfig::dm_Fuzz(pullX, pullY, fX, fY, 0.002f, 0.002f);
                    break;
                case 5:
                    // square rush
                    TransformConfig::dm_Pull(pullX, pullY, fX, fY,  -0.02f, -0.02f);
                    TransformConfig::dm_SquarePoppers(pullX, pullY, fX, fY, 30, 30, 0.08f, 0.08f);
                    break;
                case 6:
                    // alien egg zoom
                    TransformConfig::dm_AlienEggs(pullR, pullD, fX, fY, 10, 10, 1.0, 1.0);
                    TransformConfig::dm_PolarPull(pullR, fR, 1.0025f);
                    break;
                case 7:
                    // fun times!
                    pullR += sin(fX*10*3.1415)/1250;
                    pullD += cos(fY*10*3.1415)/125;
                    TransformConfig::dm_PolarPull(pullR, fR, 1.00125f);
                    break;
                case 8:
                    // sideways fun times!
                    pullD += sin((fY-fX)*5*3.1415)/200;
                    TransformConfig::dm_PolarPull(pullR, fR, 1.0025f);
                    break;
                case 9:
                    // experimental
                    pullR += sqrtf(1-sin(fR*20*3.1415)/1)/500;
                    pullR += sqrtf(1-sin(fR*20*3.1415)/1)/1000;
                    break;
                case 10:
                    // double root
                    pullX += (fX*fX-fY)/100;
                    pullY += (2*fX*fY)/100;
                    break;
                case 11:
                    // alternating lines
                    TransformConfig::dm_SlicerAdd(pullX, pullY, fX, fY, 20, 1, 0.005f, 0);
                    break;
                default:
                    break;
            }
            //*/

#endif
