// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "Sirenic.h"
#include "SirenicWnd.h"

#include "CRenderDD4.h"
#include "CRenderDX9.h"
#include "CRenderDX9Texture.h"

#include "FX_CMachineWorld.h"
#include "FX_PolarThreader.h"
#include "FX_RandomSquares.h"
#include "FX_TreeFractal.h"
#include "FX_PolyDance.h"
#include "FX_Bezier.h"

#include "Thread.h"

#include <windows.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

static const int WIDTH  = 800;
static const int HEIGHT = 600;

static const int SIRENIC_FPS = 60;
static const int SIRENIC_DELAY = (int)((1.0f / (float)SIRENIC_FPS)*1000.0f);

static Sirenic     *pSirenic = NULL;
static CRenderDD4  *pRender  = NULL;
static SirenicWnd  *pWindow  = NULL;

static void RenderThreadFunc(bool *pStopFlag, void *pContext);

LONG WINAPI UnhandledExceptionFunc(_EXCEPTION_POINTERS* ExceptionInfo)
{
    ExitProcess(0);
    return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    SetUnhandledExceptionFilter(UnhandledExceptionFunc);

    /*
    if(AllocConsole())
    {
        freopen("CONOUT$", "wt", stdout);

        SetConsoleTitle("Sirenic : Debug Console");

        printf("\n");
        printf(" Sirenic Debug Console\n");
        printf("\n");

        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);
    }
    //*/

    timeBeginPeriod(1);

    pSirenic = new Sirenic(WIDTH, HEIGHT, WIDTH);

    if(!pSirenic->Create())
    {
        MessageBox(NULL, "Failed to create sirenic object", "Sirenic", MB_OK);
        return 1;
    }

    srand(time(NULL));

    /*! MachineWorld */
    /*
    {
        FX_CMachineWorld *pMachineWorld = new FX_CMachineWorld(pSirenic, 255);

        pMachineWorld->InsertMachine(new CMachine_Grid(0.5f, 0.5f, 0.04f, 0.04f, 5, 5));

        //pMachineWorld->InsertMachine(new CMachine_Box(0.5f, 0.5f, 0.1f));

        //pMachineWorld->InsertMachine(new CMachine_Chain(0.5f, 0.60f, 0.025f, 0.0f, 10));
        //pMachineWorld->InsertMachine(new CMachine_Chain(0.5f, 0.40f, 0.025f, 0.0f, 10));

        //pMachineWorld->InsertMachine(new CMachine_Triangle(0.25f, 0.25f, 0.05f, 0.3f));
        //pMachineWorld->InsertMachine(new CMachine_Triangle(0.75f, 0.25f, 0.05f, 0.2f));

        RenderFXNode *pMachineNode = pSirenic->InsertRenderFX(pMachineWorld);
    }
    //*/

    //*
    for(float fSize = 1.0f; fSize > 0.25f; fSize -= 0.01f)
    {
        for(int r=0;r<3;r++)
        {
            float fColor = ((float)rand() / (float)RAND_MAX);

            uint08 color = (uint08)(fColor * 255.0f);

            RenderFXNode *pNode0 = pSirenic->InsertRenderFX(new FX_PolarThreader(pSirenic, fSize, color, 0.005f, 0.005f, false));
        }
    }
    //*/

    //pSirenic->InsertRenderFX(new FX_TreeFractal(pSirenic));

    pSirenic->InsertRenderFX(new FX_Bezier(pSirenic, 255, CTF_SPIN_X, CTF_SPIN_NEG_Y));
    //pSirenic->InsertRenderFX(new FX_Bezier(pSirenic, 255, CTF_SPIN_Y, CTF_SPIN_NEG_X));

    //pSirenic->InsertRenderFX(new FX_RandomSquares(pSirenic, 0.01f, 0.01f, 0.1f, 0.1f, 0.0f, 0.5f, 5));
    //pSirenic->InsertRenderFX(new FX_RandomSquares(pSirenic, 0.01f, 0.01f, 0.1f, 0.1f, 0.5f, 1.0f, 20));

    //pSirenic->InsertRenderFX(new FX_PolyDance(pSirenic, 0.01f, 0.01f, 0.01f, 0.01f, 0.01f, 1.0f, 0.0025f));

    pWindow = new SirenicWnd(WIDTH, HEIGHT, pSirenic, false);

    if(!pWindow->Create())
    {
        MessageBox(NULL, "Failed to create device window", "Sirenic", MB_OK);
        return 1;
    }

    pWindow->Show();

    pRender = new CRenderDD4(WIDTH, HEIGHT, pWindow->GetWnd());

    if(!pRender->Create())
    {
        MessageBox(NULL, "Failed to create renderer (DD4)", "Sirenic", MB_OK);
        return 1;
    }

    pWindow->SetRender(pRender);

    Thread renderThread(RenderThreadFunc, pSirenic);

    renderThread.Start();

    /*! process messages */
    while(pWindow->ProcessMessages())
    {
        Thread::Sleep(20);
    }

    renderThread.Stop();

    delete pRender;

    delete pWindow;

    delete pSirenic;

    timeEndPeriod(1);

    return 0;
}

static void RenderThreadFunc(bool *pStopFlag, void *pContext)
{
    Sirenic *pSirenic = (Sirenic*)pContext;

    int lastTime = timeGetTime();
    int skip = 0, lastFps = 0, lastDelta = 0;

    while(!*pStopFlag)
    {
        int curTime = timeGetTime();
        int delta = curTime - lastTime;

        if(delta < SIRENIC_DELAY) 
        {
            Thread::Sleep(SIRENIC_DELAY - delta); 
            curTime = timeGetTime();
        }

        int fps = (int)((1.0 / ((float)curTime - (float)lastTime)) * 1000.0);

        lastTime = curTime;

        pSirenic->DrawRenderFX();

        /*! respond to mouse movement */
        {
            float mouseX = (float)pWindow->GetMousePosX();
            float mouseY = (float)pWindow->GetMousePosY();

            mouseX /= pWindow->GetWidth();
            mouseY /= pWindow->GetHeight();

            mouseX *= WIDTH;
            mouseY *= HEIGHT;

            pSirenic->SetMousePos((uint)mouseX, (uint)mouseY);
            pSirenic->SetMouseDownL(pWindow->GetMouseDownL());

            if(pSirenic->GetMouseDownL())
            {
                //pSirenic->DrawRectWu(mouseX-8, mouseY-8, mouseX+8, mouseY+8, 1.0);
                //pSirenic->DrawRect((int)mouseX-8, (int)mouseY-8, (int)mouseX+8, (int)mouseY+8, 255);

                //pSirenic->DrawRect((int)mouseX-4, (int)mouseY-4, (int)mouseX+4, (int)mouseY+4, 255);
                //pSirenic->DrawRect((int)mouseX-2, (int)mouseY-2, (int)mouseX+2, (int)mouseY+2, 255);

                //pSirenic->DrawRect((int)mouseX-1, (int)mouseY-1, (int)mouseX+1, (int)mouseY+1, 255);

                //pSirenic->DrawString(mouseX, mouseY, "Sirenic", 64, false, 255);
            }
        }

        if(skip++%20 != 0) { fps = lastFps; delta = lastDelta; }

        lastFps = fps;
        lastDelta = delta;

        /*! render fps */
	    /*
        {
            char szBuffer[255];

            sprintf(szBuffer, "%s - ~%.02d FPS [Delay: %d ms]", "Sirenic " _SIRENIC_VERSION, fps, delta);

            pSirenic->FillRect(0, 0, WIDTH, 21, 0);
            pSirenic->DrawString(5, 16, szBuffer, 16, true, 255);
        }
	    //*/

        pSirenic->Transform();

        //*
        uint08 *pPixels = pSirenic->LockViewBuffer();

        if(!pRender->RenderPixels(pPixels))
        {
            continue;
        }

        if(!pRender->RotatePalette())
        {
            continue;
        }

        pSirenic->UnlockViewBuffer();

        pSirenic->AdvanceViewBuffer();
        //*/
    }

    return;
}
