// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "vis_sirenic.h"

#include <stdio.h>
#include <time.h>

static const int WIDTH  = 800;
static const int HEIGHT = 600;

vis_sirenic::vis_sirenic(winampVisModule *pVisModule) : pVisModule(pVisModule)
{
    pSirenic = new Sirenic(WIDTH, HEIGHT, WIDTH);

    bCreated = false;
}

vis_sirenic::~vis_sirenic()
{
    if(windowState.me) 
    {
        SetForegroundWindow(pVisModule->hwndParent);
        DestroyWindow(windowState.me);
    }

    if(pRenderDX9 != NULL)
    {
        delete pRenderDX9;
        pRenderDX9 = 0;
    }

    if(pWindow != NULL)
    {
        delete pWindow;
        pWindow = 0;
    }

    if(pSirenic != NULL)
    {
        delete pSirenic;
        pSirenic = 0;
    }
}

bool vis_sirenic::Create()
{
    if(bCreated) { return true; }

    if(!pSirenic->Create())
    {
        MessageBox(pVisModule->hwndParent, "Failed to create sirenic object", "Sirenic", MB_OK);
        return false;
    }

    pSirenic->EnableWaveform();

    srand(time(NULL));

    //
    // Polar threader!
    //
    //*
    for(float fSize = 1.0f; fSize > 0.25f; fSize -= 0.01f)
    {
        for(int r=0;r<3;r++)
        {
            float fColor = ((float)rand() / (float)RAND_MAX);

            uint08 color = (uint08)(fColor * 255.0f);

            RenderFXNode *pNode0 = pSirenic->InsertRenderFX(new FX_PolarThreader(pSirenic, fSize, color));
        }
    }
    //*/

    HWND hParentWnd = NULL;

    //
    // retrieve parent window handle
    //

    {
        HWND (*createEmbeddedWindow)(embedWindowState *v);
    
        *(void**)&createEmbeddedWindow = (void *)SendMessage(pVisModule->hwndParent, WM_WA_IPC, (LPARAM)0, IPC_GET_EMBEDIF);

        if(!createEmbeddedWindow)
        {
            return false;
        }

        windowState.flags = EMBED_FLAGS_NOTRANSPARENCY;

        windowState.r.left = 150;
        windowState.r.top = 150;
        windowState.r.right = 150 + WIDTH;
        windowState.r.bottom = 150 + HEIGHT;
    
        hParentWnd = createEmbeddedWindow(&windowState);
    }

    pWindow = new WinampRenderWnd(WIDTH, HEIGHT, hParentWnd);

    if(!pWindow->Create())
    {
        MessageBox(NULL, "Failed to create device window", "Sirenic", MB_OK);
        return false;
    }

    SetWindowText(windowState.me, pVisModule->description);

    SendMessage(pVisModule->hwndParent, WM_WA_IPC, (int)pWindow->GetWnd(), IPC_SETVISWND);

    ShowWindow(hParentWnd, SW_SHOWNA);

    pWindow->Show();

    pRenderDX9 = new CRenderDX9(WIDTH, HEIGHT, pWindow->GetWnd());

    if(!pRenderDX9->Create())
    {
        MessageBox(NULL, "Failed to create renderer (DX9)", "Sirenic", MB_OK);
        return 1;
    }

    bCreated = true;

    return true;
}

void vis_sirenic::Render()
{
    static float c = 1.0f;
    static float cs = 0.001f;

    if(c > 1.0 || c < 0.5) { cs = -cs; }

    c += cs;

    pSirenic->UpdateWaveform(pVisModule->waveformData[0]);
    pSirenic->UpdateWaveformScale(0.5f);
    pSirenic->UpdateWaveformColor(c);

    if(!pWindow->ProcessMessages())
    {
        return;
    }

    pSirenic->DrawRenderFX();

    //
    // Respond to mouse movement
    //

    {
        float mouseX = (float)pWindow->GetMousePosX();
        float mouseY = (float)pWindow->GetMousePosY();

        mouseX /= pWindow->GetWidth();
        mouseY /= pWindow->GetHeight();

        mouseX *= WIDTH;
        mouseY *= HEIGHT;

        pSirenic->SetMousePos((uint)mouseX, (uint)mouseY);
        pSirenic->SetMouseDownL(pWindow->GetMouseDownL());

        if(pSirenic->GetMouseDownL())
        {
            //pSirenic->DrawRectWu(mouseX-8, mouseY-8, mouseX+8, mouseY+8, 1.0);
            //pSirenic->DrawRect((int)mouseX-8, (int)mouseY-8, (int)mouseX+8, (int)mouseY+8, 255);

            //pSirenic->DrawRect((int)mouseX-4, (int)mouseY-4, (int)mouseX+4, (int)mouseY+4, 255);
            //pSirenic->DrawRect((int)mouseX-2, (int)mouseY-2, (int)mouseX+2, (int)mouseY+2, 255);

            //pSirenic->DrawRect((int)mouseX-1, (int)mouseY-1, (int)mouseX+1, (int)mouseY+1, 255);

            //pSirenic->DrawString(mouseX, mouseY, "Sirenic", 64, false, 255);
        }
    }

    pSirenic->Transform();

    //*
    uint08 *pPixels = pSirenic->LockViewBuffer();

    if(!pRenderDX9->RenderPixels(pPixels))
    {
        return;
    }

    if(!pRenderDX9->RotatePalette())
    {
        return;
    }

    pSirenic->UnlockViewBuffer();

    pSirenic->AdvanceViewBuffer();
    //*/

    return;
}
