// ******************************************************************
// *
// *  Sirenic is free software; you can redistribute it and/or modify
// *  it under the terms of the GNU General Public License as published
// *  by the Free Software Foundation; either version 2 of the license,
// *  or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2004 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************

#include "Sirenic.h"
#include "AutoMutex.h"
#include "TransformConfig.h"

#include <memory.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

Sirenic::Sirenic(int width, int height, int pitch, int bufferCount) : width(width), height(height), pitch(pitch), bufferCount(bufferCount)
{
    pDistortThread = 0;

    assignFunc = roAssignSimple;

    transConfig = 0;

    pRenderFXNodeH = 0;

    mouseX = mouseY = -1;

    mouseDownL = false;

    pBuffer = 0;

    pDeltaFieldFront = 0;
    pDeltaFieldBack = 0;

    pDitherMap = 0;

    viewBuffer = activeBuffer = previousBuffer = 0;

    bWaveformEnabled = false;

    activeWaveform = 0;

    waveformScale = 1.0f;
    waveformColor = 1.0f;

    bExtraSrcWeight = false;

    pBufferMutex = pRenderFXMutex = pWaveformMutex = pViewMutex = 0;

    bCreated = false;

    darkenCenter = false;
    darkenEdges = false;

    pWaveform[0] = 0;
    pWaveform[1] = 0;
}

Sirenic::~Sirenic()
{
    delete pDistortThread;

    if(pBuffer != 0)
    {
        for(int v=0;v<bufferCount;v++)
        {
            if(pBuffer[v].pMutex != 0)
            {
                delete pBuffer[v].pMutex;
                pBuffer[v].pMutex = 0;
            }

            if(pBuffer[v].pPixels != 0)
            {
                delete[] pBuffer[v].pPixels;
                pBuffer[v].pPixels = 0;
            }
        }

        delete[] pBuffer;
    }

    if(pDeltaFieldFront != 0)
    {
        delete[] pDeltaFieldFront;
        pDeltaFieldFront = 0;
    }

    if(pDeltaFieldBack != 0)
    {
        delete[] pDeltaFieldBack;
        pDeltaFieldBack = 0;
    }

    if(pDitherMap != 0)
    {
        delete[] pDitherMap;
        pDitherMap = 0;
    }

    if(pWaveform != 0)
    {
        for(uint v=0;v<2;v++)
        {
            if(pWaveform[v] != 0)
            {
                delete[] pWaveform[v];
                pWaveform[v] = 0;
            }
        }
    }

    while(pRenderFXNodeH != 0)
    {
        RemoveRenderFX(pRenderFXNodeH);
    }

    if(pBufferMutex != 0)
    {
        delete pBufferMutex;
        pBufferMutex = 0;
    }

    if(pRenderFXMutex != 0)
    {
        delete pRenderFXMutex;
        pRenderFXMutex = 0;
    }

    if(pWaveformMutex != 0)
    {
        delete pWaveformMutex;
        pWaveformMutex = 0;
    }

    if(pViewMutex != 0)
    {
        delete pViewMutex;
        pViewMutex = 0;
    }
}

bool Sirenic::Create()
{
    if(bCreated) { return true; }

    int v=0;

    pBuffer = new LockableBuffer[bufferCount];

    if(pBuffer == 0)
    {
        return false;
    }

    for(v=0;v<bufferCount;v++)
    {
        pBuffer[v].pMutex = new Mutex();

        if(pBuffer[v].pMutex == 0)
        {
            return false;
        }

        pBuffer[v].pPixels = new uint08[pitch*height];

        if(pBuffer[v].pPixels == 0)
        {
            return false;
        }

        memset(pBuffer[v].pPixels, 0, pitch*height);
    }

    activeBuffer = 0;
    viewBuffer = previousBuffer = bufferCount - 1;

    pDeltaFieldFront = new Delta[width*height];
    pDeltaFieldBack = new Delta[width*height];

    if(pDitherMap == 0)
    {
        pDitherMap = new sint08[height*width];

        for(int d=0;d<height*width;d++)
        {
            /*! @todo allow configurable dithering power */
            pDitherMap[d] = 0 - (rand()%5 - 2);
        }
    }

    for(v=0;v<2;v++)
    {
        pWaveform[v] = new uint08[576];

        if(pWaveform[v] == 0)
        {
            return false;
        }

        memset(pWaveform[v], 0, 576);
    }

    pBufferMutex = new Mutex();
    pRenderFXMutex = new Mutex();
    pWaveformMutex = new Mutex();
    pViewMutex = new Mutex();

    bCreated = true;

    srand(time(NULL));

    Distort(NULL, pDeltaFieldFront);

    pDistortThread = new Thread(DistortThreadFwd, this);

    pDistortThread->Start();

    return true;
}

void Sirenic::LinearInterpolate(float fT, CartCoord &fCoord0, CartCoord &fCoord1, CartCoord &fOut)
{
    fOut.fX = fCoord0.fX + (fCoord1.fX - fCoord0.fX)*fT;
    fOut.fY = fCoord0.fY + (fCoord1.fY - fCoord0.fY)*fT;

    return;
}

void Sirenic::DistortThread(bool *pStopFlag)
{
    while(!*pStopFlag)
    {
        Distort(pStopFlag, pDeltaFieldBack);

        Delta *pTmp = pDeltaFieldBack;

        pDeltaFieldBack  = pDeltaFieldFront;
        pDeltaFieldFront = pTmp;
    }

    return;
}

bool Sirenic::Distort(bool *pStopFlag, struct Delta *pDeltaField)
{
    if(!bCreated) { return true; }

    uint v = 0;

    TransformConfig *curTC = transConfig;

    if(transConfig == 0)
    {
        /*! randomize transformation */
        switch(rand()%4)
        {
            case 0:
                curTC = new tc_Simple();
                break;
            case 1:
                curTC = new tc_BumpyVortex();
                break;
            case 2:
                curTC = new tc_RotoZoom();
                break;
            case 3:
                curTC = new tc_CornerStretch();
                break;
            case 4:
                curTC = new tc_QuadrantRotate();
                break;
            case 5:
                curTC = new tc_PolarWaves();
                break;
            case 6:
                curTC = new tc_UnNamed();
                break;
        }
    }

    darkenCenter = true;
    darkenEdges = true;

    /*! @todo configuration for display aspect ratio */
    float fAsp = 1.333333f;

    /*! generate delta field */
    for(int y=0;y<height;y++)
    {
        for(int x=0;x<width;x++)
        {
            /*! check if we should exit this thread */
            if(pStopFlag != NULL)
            {
                if(*pStopFlag)
                {
                    return true;
                }

                static int chk = 0;

                if(chk++%100 == 0)
                {
                    Thread::Sleep(1);
                }
            }

            float fX = (((float)x / (float)width)*2.0f - 1.0f) * fAsp;
            float fY = ((float)y / (float)height)*2.0f - 1.0f;

            float fR = (float)hypot(fX, fY);
            float fD = atan2f(fY, fX);

            float pullX = 0, pullY = 0;
            float pullR = 0, pullD = 0;

            //
            // perform transformation
            //
            //*
            curTC->Transform
            (
                pullX, pullY, fX, fY,
                pullR, pullD, fR, fD
            );
            //*/

            /*! apply polar coordinate transformation */
            if(pullR != 0 || pullD != 0)
            {
                fR += pullR;
                fD += pullD;

                float nX = fR * cosf(fD);
                float nY = fR * sinf(fD);

                pullX += (fX - nX)*2;
                pullY += (fY - nY)*2;
            }

            /*! apply cartesian coordinate transformation */
            pullX *= width;
            pullY *= height;

            /*! calculate distortion */
            sint32 dstX = (sint32)(pullX + (float)x);
            sint32 dstY = (sint32)(pullY + (float)y);

            /*! fix negative rounding glitch */
            if(pullX < 0) dstX++;
            if(pullY < 0) dstY++;

            /*! calculate floating point remainder */
            float remX = pullX - (float)(int)pullX;
            float remY = pullY - (float)(int)pullY;

            /*! offset center based on remainder */
            {
                if(remX > 0.5) { dstX++; remX -= 1.0; }
                else if(remX < -0.5) { dstX--; remX += 1.0; }

                if(remY > 0.5) { dstY++; remY -= 1.0; }
                else if(remY < -0.5) { dstY--; remY += 1.0; }
            }

            /*! clamp distortion */
            {
                if(dstX < 1) { dstX = 1; }
                if(dstY < 1) { dstY = 1; }
                if(dstX > width-2) { dstX = width-2; }
                if(dstY > height-2) { dstY = height-2; }
            }

            /*! calculate offset */
            pDeltaField[v].o = dstX + dstY*pitch;

            /*! calculate weights */
            {
                pDeltaField[v].w0 = 64;
                pDeltaField[v].w1 = 64;
                pDeltaField[v].w2 = 64;
                pDeltaField[v].w3 = 64;

                if(remX > 0.0)
                {
                    pDeltaField[v].w2 += (uint08)(remX*128.0);
                    pDeltaField[v].w1 -= (uint08)(remX*128.0);
                }
                else
                {
                    pDeltaField[v].w1 += (uint08)((0.0 - remX)*128.0);
                    pDeltaField[v].w2 -= (uint08)((0.0 - remX)*128.0);
                }

                if(remY > 0.0)
                {
                    pDeltaField[v].w3 += (uint08)(remY*128.0);
                    pDeltaField[v].w0 -= (uint08)(remY*128.0);
                }
                else
                {
                    pDeltaField[v].w0 += (uint08)((0.0 - remY)*128.0);
                    pDeltaField[v].w3 -= (uint08)((0.0 - remY)*128.0);
                }
            }

            v++;
        }
    }

    if(transConfig == NULL) { delete curTC; }

    if(pStopFlag != NULL)
    {
        for(int r=0;r<200;r++)
        {
            if(*pStopFlag)
            {
                return true;
            }

            Thread::Sleep(10);
        }
    }

    return true;
}

RenderFXNode *Sirenic::InsertRenderFX(RenderFX *pRenderFX)
{
    if(!bCreated) { return 0; }

    AutoMutex threadSafe(pRenderFXMutex);

    RenderFXNode *pPreviousH = pRenderFXNodeH;

    pRenderFXNodeH = new RenderFXNode();

    pRenderFXNodeH->pPrev = 0;
    pRenderFXNodeH->pNext = pPreviousH;
    pRenderFXNodeH->pRenderFX = pRenderFX;

    if(pPreviousH != 0)
    {
        pPreviousH->pPrev = pRenderFXNodeH;
    }

    return pRenderFXNodeH;
}

bool Sirenic::RemoveRenderFX(RenderFXNode *pRenderFXNode)
{
    if(!bCreated) { return false; }

    AutoMutex threadSafe(pRenderFXMutex);

    if(pRenderFXNode->pPrev != 0)
    {
        pRenderFXNode->pPrev->pNext = pRenderFXNode->pNext;
    }
    else
    {
        pRenderFXNodeH = pRenderFXNode->pNext;
    }

    if(pRenderFXNode->pNext != 0)
    {
        pRenderFXNode->pNext->pPrev = pRenderFXNode->pPrev;
    }

    delete pRenderFXNode->pRenderFX;

    delete pRenderFXNode;

    return true;
}

bool Sirenic::DrawPoint(int x0, int y0, uint08 c)
{
    if(!bCreated) { return false; }

    if(x0 < 0) x0 = 0;
    if(x0 >= width) x0 = width-1;

    if(y0 < 0) y0 = 0;
    if(y0 >= height) y0 = height-1;

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    assignFunc(&pBuffer[previousBuffer].pPixels[y0*pitch + x0], c);

    return true;
}

bool Sirenic::DrawLine(int x0, int y0, int x1, int y1, uint08 c)
{
    if(!bCreated) { return false; }

    if(x0 < 0) x0 = 0;
    if(x0 >= width) x0 = width-1;
    if(x1 < 0) x1 = 0;
    if(x1 >= width) x1 = width-1;

    if(y0 < 0) y0 = 0;
    if(y0 >= height) y0 = height-1;
    if(y1 < 0) y1 = 0;
    if(y1 >= height) y1 = height-1;

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    Sirenic::roDrawLine(pBuffer[previousBuffer].pPixels, pitch, x0, y0, x1, y1, c, assignFunc);

    return true;
}

bool Sirenic::DrawLineWu(float x0, float y0, float x1, float y1, float c)
{
    if(!bCreated) { return false; }

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    Sirenic::roDrawLineWu(pBuffer[previousBuffer].pPixels, pitch, x0, y0, x1, y1, c);

    return true;
}

bool Sirenic::DrawBezier(CartCoord fCoord[5], float fStep, uint08 c)
{
    if(!bCreated) { return false; }

    float fCur = 0.0f;

    int v, max = (int)(1.0f / fStep);

    CartCoord fPos, fPrev, fTmp[5];

    for(v=0;v<max;v++)
    {
        LinearInterpolate(fCur, fCoord[0], fCoord[1], fTmp[0]);
        LinearInterpolate(fCur, fCoord[1], fCoord[2], fTmp[1]);
        LinearInterpolate(fCur, fCoord[2], fCoord[3], fTmp[2]);
        LinearInterpolate(fCur, fTmp[0], fTmp[1], fTmp[3]);
        LinearInterpolate(fCur, fTmp[1], fTmp[2], fTmp[4]);
        LinearInterpolate(fCur, fTmp[3], fTmp[4], fPos);

        if(v == 0) 
        {
            DrawPoint((int)fPos.fX, (int)fPos.fY, c);
        } 
        else 
        {
            DrawLine((int)fPrev.fX, (int)fPrev.fY, (int)fPos.fX, (int)fPos.fY, c);
        }

        fPrev.fX = fPos.fX;
        fPrev.fY = fPos.fY;

        fCur += fStep;
    }

    return true;
}

bool Sirenic::DrawRect(int x0, int y0, int x1, int y1, uint08 c)
{
    if(!bCreated) { return false; }

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    if(x0 < 0) x0 = 0;
    if(x0 >= width) x0 = width-1;
    if(x1 < 0) x1 = 0;
    if(x1 >= width) x1 = width-1;

    if(y0 < 0) y0 = 0;
    if(y0 >= height) y0 = height-1;
    if(y1 < 0) y1 = 0;
    if(y1 >= height) y1 = height-1;

    Sirenic::roDrawLine(pBuffer[previousBuffer].pPixels, pitch, x0, y0, x1, y0, c, assignFunc);
    Sirenic::roDrawLine(pBuffer[previousBuffer].pPixels, pitch, x1, y0, x1, y1, c, assignFunc);
    Sirenic::roDrawLine(pBuffer[previousBuffer].pPixels, pitch, x1, y1, x0, y1, c, assignFunc);
    Sirenic::roDrawLine(pBuffer[previousBuffer].pPixels, pitch, x0, y1, x0, y0, c, assignFunc);

    return true;
}

bool Sirenic::DrawRectWu(float x0, float y0, float x1, float y1, float c)
{
    if(!bCreated) { return false; }

    if(x0 < 0) x0 = 0;
    if(x0 >= (float)width) x0 = (float)width-1;
    if(x1 < 0) x1 = 0;
    if(x1 >= (float)width) x1 = (float)width-1;

    if(y0 < 0) y0 = 0;
    if(y0 >= (float)height) y0 = (float)height-1;
    if(y1 < 0) y1 = 0;
    if(y1 >= (float)height) y1 = (float)height-1;

    Sirenic::roDrawLineWu(pBuffer[previousBuffer].pPixels, pitch, x0, y0, x1, y0, c);
    Sirenic::roDrawLineWu(pBuffer[previousBuffer].pPixels, pitch, x1, y0, x1, y1, c);
    Sirenic::roDrawLineWu(pBuffer[previousBuffer].pPixels, pitch, x1, y1, x0, y1, c);
    Sirenic::roDrawLineWu(pBuffer[previousBuffer].pPixels, pitch, x0, y1, x0, y0, c);

    return true;
}

bool Sirenic::FillRect(int x0, int y0, int x1, int y1, uint08 c)
{
    if(!bCreated) { return false; }

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    if(x0 < 0) x0 = 0;
    if(x0 >= width) x0 = width-1;
    if(x1 < 0) x1 = 0;
    if(x1 >= width) x1 = width-1;

    if(y0 < 0) y0 = 0;
    if(y0 >= height) y0 = height-1;
    if(y1 < 0) y1 = 0;
    if(y1 >= height) y1 = height-1;

    Sirenic::roFillRect(pBuffer[previousBuffer].pPixels, pitch, x0, y0, x1, y1, c);

    return true;
}

#ifndef _XBOX /*! @todo support DrawString on Xbox */
bool Sirenic::DrawString(int x0, int y0, const char *str, int size, bool antiAlias, uint08 c)
{
    if(!bCreated) { return false; }

    AutoMutex activeBufferLock(pBuffer[previousBuffer].pMutex);

    Sirenic::roDrawString(pBuffer[previousBuffer].pPixels, pitch, height, x0, y0, str, size, antiAlias, c);

    return true;
}
#endif
bool Sirenic::DrawRenderFX()
{
    if(!bCreated) { return false; }

    //
    // render bitmap
    //

    if(mouseDownL)
    {
        #define BMP_W 128
        #define BMP_H 128

        static uint08 pMegaMan[] =
        {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0A, 0x0C, 0x0E, 0x0F, 0x0F, 0x0D, 0x0B, 0x09, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0x11, 0x1E, 0x2D, 0x36, 0x3B, 0x37, 0x2F, 0x24, 0x16, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x1E, 0x39, 0x58, 0x6E, 0x75, 0x70, 0x5F, 0x45, 0x28, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x28, 0x4F, 0x79, 0x99, 0xA7, 0xA3, 0x8E, 0x6C, 0x42, 0x1E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x02, 0x03, 0x03, 0x02, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x03, 0x02, 0x03, 0x03, 0x03, 0x03, 0x03, 0x0E, 0x29, 0x4F, 0x7A, 0x9C, 0xB5, 0xBA, 0xAE, 0x90, 0x65, 0x3B, 0x19, 0x0C, 0x0A, 0x07, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0C, 0x0E, 0x0E, 0x0E, 0x0E, 0x0C, 0x07, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0C, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x12, 0x22, 0x3B, 0x57, 0x75, 0x97, 0xAE, 0xB7, 0xAB, 0x8D, 0x68, 0x43, 0x2D, 0x1F, 0x12, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0E, 0x18, 0x1D, 0x1C, 0x1C, 0x1D, 0x18, 0x0E, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x0E, 0x18, 0x1D, 0x1D, 0x1D, 0x1D, 0x1C, 0x1D, 0x1D, 0x1D, 0x1B, 0x1A, 0x20, 0x2C, 0x41, 0x66, 0x8B, 0xAB, 0xB7, 0xAF, 0x99, 0x77, 0x58, 0x3A, 0x1E, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x17, 0x25, 0x2C, 0x2C, 0x2C, 0x2C, 0x26, 0x17, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x17, 0x25, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x25, 0x18, 0x0E, 0x0C, 0x18, 0x3A, 0x64, 0x90, 0xAE, 0xBA, 0xB5, 0x9D, 0x7A, 0x4F, 0x29, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x08, 0x13, 0x21, 0x30, 0x33, 0x32, 0x32, 0x32, 0x31, 0x22, 0x13, 0x08, 0x03, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x07, 0x13, 0x21, 0x30, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x30, 0x1D, 0x0D, 0x02, 0x05, 0x1D, 0x41, 0x6B, 0x8E, 0xA3, 0xA6, 0x97, 0x78, 0x4D, 0x27, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x08, 0x0D, 0x12, 0x1C, 0x28, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x2A, 0x1F, 0x15, 0x0F, 0x0A, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x08, 0x0C, 0x12, 0x1C, 0x28, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x24, 0x16, 0x0D, 0x09, 0x12, 0x2A, 0x45, 0x5D, 0x6C, 0x6D, 0x64, 0x4D, 0x2F, 0x16, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x08, 0x11, 0x18, 0x1F, 0x26, 0x2D, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x30, 0x2A, 0x25, 0x1E, 0x14, 0x0A, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x08, 0x10, 0x18, 0x1F, 0x25, 0x2C, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x2B, 0x21, 0x1B, 0x16, 0x14, 0x1A, 0x25, 0x2E, 0x32, 0x31, 0x29, 0x1E, 0x0F, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0C, 0x19, 0x25, 0x2D, 0x2E, 0x30, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x31, 0x2A, 0x1B, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0C, 0x19, 0x25, 0x2C, 0x2F, 0x30, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x33, 0x32, 0x30, 0x2E, 0x2C, 0x26, 0x1A, 0x11, 0x0D, 0x0D, 0x0C, 0x0B, 0x08, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1F, 0x31, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x32, 0x31, 0x31, 0x29, 0x1B, 0x0D, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0F, 0x1F, 0x30, 0x32, 0x32, 0x32, 0x33, 0x32, 0x32, 0x33, 0x32, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x30, 0x1D, 0x0D, 0x01, 0x00, 0x00, 0x00, 0x04, 0x08, 0x0B, 0x0C, 0x0D, 0x0C, 0x0B, 0x08, 0x04, 0x02, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x11, 0x23, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x32, 0x32, 0x31, 0x2D, 0x29, 0x25, 0x1E, 0x13, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0F, 0x1F, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x31, 0x1C, 0x0A, 0x00, 0x00, 0x04, 0x0F, 0x1E, 0x2A, 0x33, 0x38, 0x36, 0x2E, 0x21, 0x18, 0x10, 0x0D, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x11, 0x23, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2E, 0x26, 0x1D, 0x14, 0x0F, 0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0C, 0x1C, 0x2F, 0x32, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x32, 0x32, 0x32, 0x32, 0x33, 0x32, 0x32, 0x32, 0x2F, 0x1A, 0x08, 0x00, 0x01, 0x15, 0x2F, 0x4C, 0x65, 0x74, 0x77, 0x6D, 0x59, 0x40, 0x2C, 0x1F, 0x19, 0x11, 0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x10, 0x1F, 0x31, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2C, 0x1F, 0x12, 0x08, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0C, 0x1B, 0x2C, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x19, 0x09, 0x00, 0x0A, 0x25, 0x4D, 0x77, 0x98, 0xAA, 0xA9, 0x98, 0x7C, 0x59, 0x3F, 0x2F, 0x26, 0x1A, 0x0C, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0C, 0x18, 0x25, 0x2D, 0x2E, 0x30, 0x32, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x1A, 0x0B, 0x00, 0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0A, 0x0C, 0x0E, 0x14, 0x20, 0x2C, 0x33, 0x33, 0x32, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x2C, 0x1C, 0x0E, 0x05, 0x0B, 0x27, 0x4E, 0x78, 0x9A, 0xAB, 0xAA, 0x98, 0x7C, 0x5A, 0x42, 0x35, 0x31, 0x21, 0x10, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x09, 0x10, 0x18, 0x1F, 0x25, 0x2D, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2A, 0x17, 0x07, 0x00, 0x00, 0x00, 0x07, 0x0C, 0x0E, 0x09, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x10, 0x1E, 0x28, 0x2A, 0x27, 0x27, 0x2F, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x2E, 0x22, 0x17, 0x10, 0x0C, 0x1B, 0x36, 0x56, 0x6E, 0x7A, 0x7A, 0x6D, 0x58, 0x41, 0x3A, 0x33, 0x33, 0x25, 0x12, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x08, 0x0C, 0x12, 0x1C, 0x28, 0x33, 0x33, 0x33, 0x33, 0x32, 0x2A, 0x17, 0x05, 0x00, 0x00, 0x02, 0x0E, 0x1A, 0x1B, 0x12, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x1E, 0x3A, 0x4F, 0x51, 0x43, 0x35, 0x30, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x30, 0x28, 0x22, 0x1D, 0x18, 0x13, 0x1C, 0x2B, 0x36, 0x3C, 0x3B, 0x36, 0x2B, 0x23, 0x2B, 0x32, 0x32, 0x25, 0x12, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x07, 0x13, 0x21, 0x31, 0x33, 0x33, 0x32, 0x33, 0x2A, 0x17, 0x08, 0x00, 0x00, 0x08, 0x1A, 0x29, 0x2A, 0x1D, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2C, 0x57, 0x78, 0x7B, 0x61, 0x42, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x30, 0x2E, 0x2C, 0x26, 0x18, 0x0E, 0x0C, 0x0D, 0x0F, 0x0F, 0x0E, 0x0C, 0x14, 0x24, 0x31, 0x31, 0x23, 0x11, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x17, 0x25, 0x2D, 0x2E, 0x30, 0x33, 0x2C, 0x1B, 0x0C, 0x04, 0x05, 0x10, 0x23, 0x32, 0x33, 0x28, 0x16, 0x08, 0x03, 0x02, 0x00, 0x00, 0x0E, 0x36, 0x6D, 0x97, 0x9B, 0x77, 0x4C, 0x35, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x34, 0x35, 0x36, 0x36, 0x37, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x38, 0x36, 0x24, 0x12, 0x05, 0x01, 0x01, 0x01, 0x01, 0x06, 0x14, 0x26, 0x35, 0x35, 0x26, 0x14, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0F, 0x18, 0x1F, 0x25, 0x2D, 0x33, 0x32, 0x24, 0x17, 0x0F, 0x10, 0x1A, 0x2A, 0x33, 0x32, 0x2E, 0x1F, 0x12, 0x0C, 0x07, 0x02, 0x00, 0x0E, 0x39, 0x71, 0x9E, 0xA2, 0x7D, 0x4F, 0x35, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x39, 0x46, 0x52, 0x55, 0x57, 0x58, 0x5B, 0x5E, 0x61, 0x65, 0x67, 0x66, 0x66, 0x66, 0x5A, 0x3D, 0x20, 0x0F, 0x0B, 0x0B, 0x0C, 0x0E, 0x11, 0x25, 0x3F, 0x55, 0x55, 0x3D, 0x20, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0C, 0x12, 0x1C, 0x29, 0x32, 0x33, 0x2D, 0x24, 0x1D, 0x1D, 0x24, 0x2E, 0x33, 0x33, 0x30, 0x27, 0x1F, 0x19, 0x0F, 0x06, 0x00, 0x0E, 0x35, 0x6B, 0x96, 0x9A, 0x78, 0x4C, 0x35, 0x33, 0x32, 0x32, 0x33, 0x32, 0x36, 0x4B, 0x6F, 0x8C, 0x93, 0x94, 0x96, 0x9B, 0xA2, 0xAA, 0xB1, 0xB4, 0xB5, 0xB4, 0xB5, 0x9C, 0x65, 0x34, 0x1E, 0x1B, 0x1D, 0x1E, 0x20, 0x24, 0x41, 0x6D, 0x91, 0x90, 0x67, 0x35, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x07, 0x13, 0x21, 0x30, 0x32, 0x33, 0x2F, 0x2C, 0x2D, 0x2F, 0x30, 0x33, 0x33, 0x31, 0x2F, 0x2D, 0x26, 0x16, 0x0A, 0x01, 0x0C, 0x32, 0x64, 0x8C, 0x90, 0x71, 0x49, 0x34, 0x33, 0x33, 0x33, 0x33, 0x33, 0x40, 0x6B, 0xA4, 0xD1, 0xDA, 0xDC, 0xDD, 0xE1, 0xE7, 0xED, 0xF3, 0xF6, 0xF5, 0xF5, 0xF6, 0xD3, 0x8B, 0x4D, 0x2F, 0x2D, 0x2D, 0x2F, 0x30, 0x35, 0x5D, 0x9F, 0xD7, 0xD4, 0x98, 0x4C, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x17, 0x25, 0x2E, 0x31, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x30, 0x1B, 0x0B, 0x01, 0x0C, 0x31, 0x63, 0x8C, 0x90, 0x6F, 0x47, 0x34, 0x33, 0x33, 0x33, 0x34, 0x3B, 0x5F, 0x97, 0xD8, 0xFA, 0xFB, 0xFD, 0xFE, 0xFF, 0xFF, 0xFE, 0xF5, 0xF3, 0xF3, 0xF3, 0xF4, 0xD8, 0x9B, 0x64, 0x40, 0x37, 0x35, 0x33, 0x30, 0x36, 0x69, 0xC2, 0xFF, 0xFD, 0xBE, 0x5D, 0x15, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0E, 0x17, 0x20, 0x28, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x1C, 0x0A, 0x00, 0x08, 0x2E, 0x67, 0x95, 0x99, 0x72, 0x46, 0x33, 0x33, 0x34, 0x3B, 0x48, 0x60, 0x8A, 0xBC, 0xED, 0xFB, 0xFC, 0xFD, 0xFF, 0xFE, 0xF7, 0xD6, 0xB9, 0xAE, 0xAD, 0xAE, 0xAE, 0xA2, 0x8D, 0x76, 0x5F, 0x51, 0x3F, 0x2C, 0x24, 0x2A, 0x62, 0xC4, 0xFF, 0xFE, 0xC6, 0x61, 0x16, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x08, 0x0B, 0x12, 0x1E, 0x2B, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x30, 0x1B, 0x09, 0x00, 0x05, 0x2E, 0x6A, 0x9C, 0xA1, 0x74, 0x45, 0x34, 0x32, 0x38, 0x50, 0x73, 0x98, 0xB8, 0xD8, 0xF5, 0xFC, 0xFD, 0xFE, 0xFF, 0xFE, 0xC8, 0x94, 0x66, 0x55, 0x55, 0x56, 0x56, 0x5C, 0x70, 0x84, 0x8D, 0x7F, 0x56, 0x28, 0x13, 0x19, 0x52, 0xB3, 0xFF, 0xFE, 0xBD, 0x5C, 0x16, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x08, 0x13, 0x21, 0x31, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x2C, 0x1A, 0x09, 0x00, 0x08, 0x2F, 0x67, 0x95, 0x99, 0x6C, 0x40, 0x33, 0x34, 0x3F, 0x6B, 0xA8, 0xD5, 0xE6, 0xF1, 0xF9, 0xFF, 0xFF, 0xFF, 0xFF, 0xE3, 0x9D, 0x59, 0x25, 0x14, 0x14, 0x14, 0x14, 0x22, 0x55, 0x92, 0xBF, 0xB8, 0x77, 0x2E, 0x09, 0x0D, 0x43, 0xA3, 0xF5, 0xF8, 0xB1, 0x58, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0A, 0x17, 0x26, 0x2D, 0x2E, 0x31, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2D, 0x1F, 0x13, 0x0D, 0x15, 0x34, 0x5C, 0x79, 0x79, 0x57, 0x38, 0x32, 0x35, 0x49, 0x83, 0xD0, 0xFA, 0xFB, 0xFC, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xD9, 0x86, 0x3B, 0x08, 0x00, 0x00, 0x00, 0x00, 0x08, 0x47, 0x9F, 0xE5, 0xE3, 0x92, 0x35, 0x07, 0x08, 0x3B, 0x9B, 0xF1, 0xF6, 0xB0, 0x57, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x02, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x02, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0F, 0x18, 0x1F, 0x26, 0x2E, 0x33, 0x33, 0x36, 0x3A, 0x3D, 0x3E, 0x3B, 0x35, 0x2F, 0x2D, 0x3C, 0x4C, 0x53, 0x50, 0x3C, 0x2F, 0x30, 0x35, 0x4D, 0x8A, 0xD9, 0xF9, 0xFB, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xD6, 0x7E, 0x2E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3D, 0x9E, 0xEB, 0xEA, 0x93, 0x2F, 0x00, 0x00, 0x32, 0x95, 0xF1, 0xF8, 0xB3, 0x5A, 0x15, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x09, 0x0D, 0x0F, 0x0E, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0C, 0x07, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x07, 0x0B, 0x12, 0x1D, 0x29, 0x33, 0x34, 0x3B, 0x47, 0x55, 0x62, 0x68, 0x66, 0x5B, 0x4F, 0x45, 0x3C, 0x31, 0x2A, 0x23, 0x26, 0x2F, 0x36, 0x4F, 0x88, 0xCF, 0xFA, 0xFB, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xD6, 0x7E, 0x2D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x39, 0x97, 0xE3, 0xE1, 0x88, 0x27, 0x00, 0x00, 0x2D, 0x92, 0xF1, 0xFA, 0xB7, 0x5B, 0x16, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x09, 0x12, 0x1A, 0x1E, 0x1E, 0x1E, 0x1F, 0x1E, 0x1E, 0x1E, 0x1E, 0x1A, 0x0F, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x03, 0x08, 0x13, 0x23, 0x32, 0x35, 0x41, 0x58, 0x73, 0x8C, 0x9B, 0x9B, 0x8C, 0x70, 0x4E, 0x2F, 0x15, 0x0C, 0x11, 0x1F, 0x2D, 0x36, 0x51, 0x83, 0xC3, 0xF3, 0xFB, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xDA, 0x86, 0x3B, 0x08, 0x00, 0x00, 0x00, 0x00, 0x08, 0x41, 0x95, 0xD8, 0xD4, 0x83, 0x2A, 0x00, 0x00, 0x33, 0x96, 0xF1, 0xF9, 0xB4, 0x5A, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0D, 0x1A, 0x27, 0x2D, 0x2D, 0x2D, 0x2D, 0x2E, 0x2D, 0x2D, 0x2E, 0x26, 0x18, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x19, 0x26, 0x30, 0x43, 0x64, 0x8D, 0xAF, 0xC3, 0xC5, 0xB0, 0x88, 0x53, 0x24, 0x05, 0x00, 0x09, 0x1A, 0x2D, 0x37, 0x54, 0x84, 0xC3, 0xF1, 0xFB, 0xFB, 0xFA, 0xFC, 0xFD, 0xFD, 0xFF, 0xE0, 0x9B, 0x58, 0x25, 0x15, 0x14, 0x14, 0x15, 0x25, 0x5E, 0xA4, 0xDA, 0xD7, 0x92, 0x44, 0x17, 0x1B, 0x50, 0xA8, 0xF4, 0xF2, 0xA9, 0x52, 0x12, 0x00, 0x02, 0x08, 0x0B, 0x0D, 0x0E, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0B, 0x08, 0x05, 0x00, 0x04, 0x10, 0x1F, 0x30, 0x36, 0x36, 0x36, 0x36, 0x35, 0x36, 0x36, 0x35, 0x31, 0x24, 0x17, 0x10, 0x0C, 0x08, 0x05, 0x01, 0x01, 0x06, 0x0A, 0x0C, 0x0D, 0x0A, 0x07, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x10, 0x19, 0x22, 0x3F, 0x6A, 0x9B, 0xC2, 0xD7, 0xD4, 0xBA, 0x8D, 0x50, 0x1D, 0x00, 0x00, 0x08, 0x1B, 0x2F, 0x38, 0x57, 0x8D, 0xCA, 0xF4, 0xFB, 0xF7, 0xF2, 0xF4, 0xF7, 0xFB, 0xFF, 0xED, 0xBD, 0x8E, 0x66, 0x57, 0x56, 0x56, 0x56, 0x66, 0x94, 0xC7, 0xEA, 0xE6, 0xB4, 0x75, 0x4C, 0x51, 0x82, 0xCC, 0xFE, 0xF1, 0x9F, 0x48, 0x0D, 0x00, 0x04, 0x13, 0x23, 0x2E, 0x30, 0x2E, 0x2B, 0x2A, 0x2D, 0x2F, 0x2D, 0x24, 0x16, 0x0B, 0x01, 0x04, 0x13, 0x27, 0x3F, 0x4D, 0x4F, 0x4D, 0x4B, 0x49, 0x4A, 0x4A, 0x4A, 0x47, 0x42, 0x3E, 0x37, 0x2C, 0x1B, 0x0C, 0x02, 0x03, 0x0F, 0x21, 0x2F, 0x30, 0x24, 0x14, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x08, 0x0C, 0x16, 0x39, 0x68, 0x9B, 0xC3, 0xD4, 0xCE, 0xB2, 0x84, 0x4B, 0x1B, 0x00, 0x00, 0x09, 0x1D, 0x33, 0x37, 0x59, 0x91, 0xD1, 0xF6, 0xF9, 0xF3, 0xEA, 0xEC, 0xF0, 0xF7, 0xFE, 0xFA, 0xE4, 0xCC, 0xB8, 0xB0, 0xAF, 0xAF, 0xAF, 0xB9, 0xD0, 0xE9, 0xF6, 0xF2, 0xD8, 0xAF, 0x96, 0x99, 0xBD, 0xF1, 0xFD, 0xEA, 0x93, 0x3C, 0x09, 0x00, 0x07, 0x23, 0x45, 0x5B, 0x60, 0x5B, 0x54, 0x54, 0x5A, 0x5E, 0x5B, 0x47, 0x29, 0x11, 0x02, 0x04, 0x16, 0x38, 0x5D, 0x76, 0x7B, 0x76, 0x6E, 0x6D, 0x6D, 0x6D, 0x6D, 0x6C, 0x71, 0x74, 0x6C, 0x56, 0x32, 0x14, 0x04, 0x06, 0x1A, 0x3D, 0x5E, 0x5F, 0x45, 0x24, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x0F, 0x2F, 0x5C, 0x8C, 0xB0, 0xC1, 0xBD, 0xA5, 0x7E, 0x4B, 0x21, 0x03, 0x00, 0x0B, 0x1D, 0x31, 0x38, 0x57, 0x8D, 0xCC, 0xF3, 0xF9, 0xED, 0xE3, 0xE4, 0xEB, 0xF4, 0xFD, 0xFF, 0xFC, 0xF8, 0xF6, 0xF4, 0xF4, 0xF5, 0xF4, 0xF4, 0xF3, 0xF0, 0xEF, 0xEC, 0xEA, 0xE4, 0xE0, 0xE1, 0xEF, 0xFD, 0xFC, 0xD4, 0x80, 0x34, 0x09, 0x00, 0x11, 0x37, 0x6A, 0x8B, 0x92, 0x8A, 0x81, 0x80, 0x89, 0x91, 0x89, 0x6D, 0x3E, 0x18, 0x04, 0x05, 0x1B, 0x47, 0x78, 0x99, 0xA0, 0x99, 0x8E, 0x8D, 0x8D, 0x8C, 0x8C, 0x8F, 0x98, 0x9E, 0x97, 0x76, 0x45, 0x1A, 0x05, 0x06, 0x23, 0x55, 0x82, 0x84, 0x5F, 0x31, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x24, 0x48, 0x6E, 0x8E, 0xA5, 0xAB, 0x9E, 0x83, 0x5C, 0x35, 0x16, 0x0B, 0x0E, 0x19, 0x26, 0x30, 0x4B, 0x7C, 0xB6, 0xE0, 0xEA, 0xE3, 0xDB, 0xE2, 0xEA, 0xF5, 0xFC, 0xFD, 0xFB, 0xF8, 0xF4, 0xF5, 0xF5, 0xF4, 0xF4, 0xEF, 0xDF, 0xCE, 0xC3, 0xC7, 0xDE, 0xFB, 0xFF, 0xFF, 0xFF, 0xED, 0xD3, 0x9F, 0x62, 0x2E, 0x11, 0x0F, 0x26, 0x54, 0x89, 0xAC, 0xB3, 0xAC, 0xA1, 0xA0, 0xAC, 0xB6, 0xAD, 0x88, 0x4F, 0x1F, 0x06, 0x05, 0x1B, 0x46, 0x78, 0x99, 0xA0, 0x98, 0x8E, 0x8C, 0x8C, 0x8C, 0x8C, 0x8E, 0x99, 0xA0, 0x9B, 0x7A, 0x47, 0x1B, 0x05, 0x07, 0x24, 0x57, 0x85, 0x88, 0x62, 0x32, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x19, 0x30, 0x49, 0x62, 0x81, 0x97, 0xA0, 0x96, 0x7C, 0x5A, 0x38, 0x23, 0x17, 0x12, 0x17, 0x1D, 0x36, 0x66, 0x9E, 0xCA, 0xD8, 0xD6, 0xD7, 0xE2, 0xF1, 0xFC, 0xFE, 0xFA, 0xE6, 0xCE, 0xB8, 0xAF, 0xAF, 0xAF, 0xAF, 0xA9, 0x9A, 0x8A, 0x7F, 0x8C, 0xBD, 0xF7, 0xFF, 0xFF, 0xF2, 0xBA, 0x87, 0x5E, 0x3F, 0x2C, 0x26, 0x2F, 0x49, 0x6E, 0x97, 0xB1, 0xB6, 0xAF, 0xA8, 0xA9, 0xB5, 0xBE, 0xB6, 0x8F, 0x56, 0x23, 0x08, 0x04, 0x15, 0x35, 0x5B, 0x73, 0x7A, 0x73, 0x6C, 0x6B, 0x6B, 0x6B, 0x6B, 0x6C, 0x75, 0x7D, 0x79, 0x60, 0x38, 0x16, 0x04, 0x06, 0x1C, 0x45, 0x6A, 0x6C, 0x4E, 0x28, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x0E, 0x19, 0x25, 0x38, 0x5C, 0x80, 0x9D, 0xA9, 0xA0, 0x86, 0x64, 0x45, 0x25, 0x0E, 0x09, 0x0D, 0x22, 0x4F, 0x87, 0xB3, 0xC6, 0xCC, 0xD4, 0xE5, 0xF9, 0xFC, 0xFD, 0xF3, 0xC4, 0x91, 0x67, 0x57, 0x56, 0x56, 0x56, 0x53, 0x4C, 0x44, 0x3E, 0x52, 0x91, 0xDF, 0xFE, 0xFD, 0xD2, 0x80, 0x43, 0x29, 0x24, 0x31, 0x47, 0x5C, 0x71, 0x84, 0x96, 0xA4, 0xA7, 0xA3, 0x9F, 0xA0, 0xAB, 0xB5, 0xAD, 0x89, 0x54, 0x26, 0x0A, 0x03, 0x0F, 0x24, 0x3B, 0x48, 0x4B, 0x48, 0x45, 0x44, 0x45, 0x44, 0x45, 0x46, 0x4B, 0x4F, 0x4C, 0x43, 0x27, 0x10, 0x04, 0x03, 0x14, 0x30, 0x48, 0x48, 0x36, 0x1C, 0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x06, 0x08, 0x0C, 0x18, 0x3A, 0x63, 0x8E, 0xAA, 0xB3, 0xAA, 0x90, 0x6A, 0x3C, 0x15, 0x01, 0x03, 0x16, 0x3B, 0x6D, 0x97, 0xB0, 0xC0, 0xD3, 0xE6, 0xF9, 0xFC, 0xFC, 0xDA, 0x97, 0x56, 0x25, 0x14, 0x14, 0x14, 0x14, 0x15, 0x13, 0x12, 0x11, 0x23, 0x62, 0xB4, 0xEE, 0xEB, 0xA6, 0x50, 0x17, 0x0B, 0x18, 0x3D, 0x6A, 0x8C, 0x97, 0x96, 0x92, 0x95, 0x96, 0x95, 0x93, 0x96, 0x9F, 0xA8, 0xA0, 0x80, 0x51, 0x28, 0x0B, 0x02, 0x0B, 0x1A, 0x2D, 0x34, 0x34, 0x34, 0x33, 0x33, 0x33, 0x33, 0x34, 0x33, 0x35, 0x36, 0x36, 0x30, 0x1D, 0x0C, 0x01, 0x01, 0x0E, 0x23, 0x35, 0x35, 0x27, 0x14, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x1E, 0x42, 0x6C, 0x90, 0xAA, 0xB4, 0xAA, 0x8D, 0x5B, 0x2C, 0x11, 0x0B, 0x14, 0x2B, 0x4D, 0x6F, 0x8F, 0xB2, 0xD3, 0xE7, 0xED, 0xE6, 0xCF, 0xA1, 0x64, 0x2C, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x39, 0x7E, 0xB6, 0xB6, 0x7B, 0x35, 0x0B, 0x08, 0x22, 0x53, 0x8B, 0xAE, 0xB3, 0xA5, 0x97, 0x92, 0x91, 0x93, 0x93, 0x96, 0xA0, 0xA8, 0xA0, 0x80, 0x54, 0x2A, 0x0C, 0x01, 0x0A, 0x1A, 0x2D, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x2F, 0x2E, 0x26, 0x16, 0x08, 0x00, 0x01, 0x0A, 0x1C, 0x2A, 0x2B, 0x1F, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x28, 0x45, 0x65, 0x88, 0xA0, 0xA9, 0x9C, 0x79, 0x51, 0x32, 0x21, 0x1A, 0x1C, 0x28, 0x41, 0x70, 0xAD, 0xE3, 0xEF, 0xDB, 0xB2, 0x82, 0x5A, 0x35, 0x13, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1E, 0x49, 0x6C, 0x71, 0x57, 0x32, 0x1F, 0x29, 0x48, 0x74, 0xA2, 0xBD, 0xBE, 0xB0, 0xA2, 0x9D, 0x9C, 0x9D, 0x9E, 0x9F, 0xAB, 0xB5, 0xAD, 0x89, 0x59, 0x2D, 0x0D, 0x02, 0x0C, 0x1D, 0x31, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x2F, 0x27, 0x21, 0x1A, 0x0F, 0x05, 0x00, 0x00, 0x07, 0x13, 0x1C, 0x1D, 0x15, 0x0A, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x15, 0x24, 0x38, 0x5B, 0x7C, 0x96, 0x9E, 0x93, 0x7B, 0x5E, 0x43, 0x27, 0x13, 0x0F, 0x1F, 0x57, 0xAA, 0xEE, 0xF3, 0xC5, 0x7D, 0x3E, 0x20, 0x13, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0F, 0x21, 0x2F, 0x36, 0x3C, 0x40, 0x4D, 0x63, 0x7C, 0x95, 0xAA, 0xB9, 0xB9, 0xB2, 0xA9, 0xA5, 0xA5, 0xA6, 0xA6, 0xA9, 0xB5, 0xBE, 0xB5, 0x90, 0x5D, 0x2F, 0x0D, 0x01, 0x0C, 0x1E, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x29, 0x1D, 0x12, 0x0C, 0x08, 0x02, 0x00, 0x00, 0x03, 0x0A, 0x0E, 0x0E, 0x0A, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0x0B, 0x17, 0x37, 0x5D, 0x84, 0x9F, 0xAA, 0xA3, 0x8D, 0x6A, 0x3A, 0x15, 0x02, 0x0F, 0x46, 0x9E, 0xE5, 0xE8, 0xAA, 0x57, 0x19, 0x06, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x09, 0x0C, 0x13, 0x2D, 0x50, 0x77, 0x97, 0xA3, 0xA2, 0xA0, 0xA3, 0xA3, 0xA1, 0xA0, 0x9F, 0x9E, 0x9F, 0x9F, 0xA1, 0xAC, 0xB6, 0xAD, 0x8A, 0x59, 0x2C, 0x0C, 0x02, 0x0B, 0x1C, 0x2F, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x31, 0x22, 0x13, 0x08, 0x03, 0x02, 0x00, 0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x23, 0x4E, 0x7D, 0xA5, 0xBD, 0xC0, 0xAF, 0x8A, 0x52, 0x23, 0x0A, 0x0E, 0x39, 0x7D, 0xB7, 0xB9, 0x89, 0x48, 0x17, 0x08, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x03, 0x0E, 0x2A, 0x52, 0x7B, 0x98, 0x9C, 0x8F, 0x81, 0x7D, 0x7D, 0x7E, 0x7E, 0x7E, 0x7F, 0x7F, 0x7F, 0x80, 0x88, 0x91, 0x89, 0x6E, 0x48, 0x24, 0x0B, 0x01, 0x09, 0x16, 0x25, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x2B, 0x2C, 0x2B, 0x26, 0x17, 0x0A, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x22, 0x50, 0x85, 0xB1, 0xC5, 0xC5, 0xB4, 0x95, 0x6B, 0x42, 0x25, 0x1E, 0x31, 0x53, 0x70, 0x76, 0x6A, 0x55, 0x3B, 0x2E, 0x22, 0x11, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x10, 0x18, 0x23, 0x37, 0x4B, 0x5E, 0x6D, 0x6F, 0x62, 0x56, 0x51, 0x50, 0x51, 0x52, 0x52, 0x52, 0x52, 0x53, 0x54, 0x59, 0x5E, 0x5A, 0x48, 0x2F, 0x19, 0x09, 0x00, 0x06, 0x0F, 0x18, 0x1D, 0x1C, 0x1C, 0x1C, 0x1C, 0x1D, 0x1C, 0x1C, 0x18, 0x0E, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0x28, 0x58, 0x8F, 0xB7, 0xC3, 0xBA, 0xA8, 0x95, 0x80, 0x6A, 0x55, 0x41, 0x33, 0x2F, 0x31, 0x3A, 0x54, 0x6F, 0x7D, 0x75, 0x52, 0x26, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x13, 0x2D, 0x42, 0x4C, 0x4B, 0x41, 0x37, 0x36, 0x37, 0x31, 0x2C, 0x2A, 0x29, 0x29, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2D, 0x2F, 0x2D, 0x25, 0x1A, 0x0F, 0x06, 0x00, 0x03, 0x08, 0x0C, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0C, 0x07, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2C, 0x59, 0x8B, 0xAC, 0xB3, 0xA8, 0x99, 0x91, 0x93, 0x94, 0x8A, 0x6A, 0x3E, 0x1B, 0x0C, 0x16, 0x44, 0x89, 0xC1, 0xBE, 0x87, 0x40, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x26, 0x52, 0x76, 0x7A, 0x5F, 0x37, 0x17, 0x0E, 0x0E, 0x0E, 0x0D, 0x0C, 0x0C, 0x0D, 0x0C, 0x0D, 0x0C, 0x0C, 0x0D, 0x0D, 0x0D, 0x0D, 0x0D, 0x0B, 0x09, 0x06, 0x02, 0x00, 0x00, 0x02, 0x03, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x03, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x25, 0x48, 0x6F, 0x8B, 0x97, 0x94, 0x91, 0x97, 0xA6, 0xB2, 0xAC, 0x89, 0x4F, 0x1E, 0x07, 0x0B, 0x3E, 0x91, 0xDB, 0xE0, 0xA5, 0x57, 0x1D, 0x0D, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x18, 0x41, 0x75, 0x9B, 0x95, 0x68, 0x30, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x1A, 0x2F, 0x48, 0x5F, 0x73, 0x7F, 0x8A, 0x9B, 0xAB, 0xB6, 0xB0, 0x94, 0x63, 0x35, 0x1B, 0x20, 0x4B, 0x90, 0xCD, 0xD2, 0xA4, 0x69, 0x38, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x36, 0x5D, 0x8A, 0xA7, 0x99, 0x65, 0x2C, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x0E, 0x1A, 0x26, 0x36, 0x51, 0x6A, 0x86, 0x9D, 0xAA, 0xAD, 0xA5, 0x94, 0x76, 0x5A, 0x47, 0x4B, 0x67, 0x8E, 0xAE, 0xB1, 0x99, 0x78, 0x5D, 0x52, 0x52, 0x53, 0x53, 0x52, 0x52, 0x53, 0x53, 0x5C, 0x79, 0x97, 0xA4, 0x8D, 0x5D, 0x2B, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x06, 0x09, 0x0B, 0x17, 0x34, 0x59, 0x81, 0x9E, 0xA6, 0xA1, 0x97, 0x90, 0x89, 0x82, 0x7C, 0x7E, 0x86, 0x8F, 0x96, 0x98, 0x91, 0x89, 0x81, 0x7E, 0x7E, 0x7E, 0x7F, 0x7E, 0x7E, 0x7E, 0x7F, 0x83, 0x92, 0xA1, 0x9F, 0x81, 0x55, 0x2A, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x23, 0x4D, 0x7C, 0xA0, 0xA7, 0x9D, 0x93, 0x93, 0x97, 0x9C, 0x9F, 0x9E, 0x99, 0x94, 0x90, 0x8F, 0x93, 0x98, 0x9D, 0x9E, 0x9F, 0x9F, 0x9F, 0x9F, 0x9F, 0x9F, 0x9F, 0xA0, 0xA6, 0xAB, 0xA0, 0x7E, 0x51, 0x2A, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x4B, 0x7D, 0xA0, 0xA7, 0x9C, 0x93, 0x93, 0x9A, 0xA2, 0xA6, 0xA5, 0x9D, 0x95, 0x8E, 0x8E, 0x94, 0x9C, 0xA3, 0xA6, 0xA6, 0xA6, 0xA6, 0xA6, 0xA6, 0xA6, 0xA6, 0xA6, 0xAB, 0xAD, 0xA0, 0x7E, 0x50, 0x29, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x24, 0x4E, 0x7D, 0xA0, 0xA8, 0x9C, 0x93, 0x92, 0x97, 0x9B, 0x9E, 0x9D, 0x99, 0x93, 0x90, 0x90, 0x92, 0x97, 0x9C, 0x9D, 0x9E, 0x9E, 0x9E, 0x9E, 0x9D, 0x9E, 0x9E, 0x9F, 0xA6, 0xAA, 0xA0, 0x7E, 0x51, 0x2A, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x29, 0x52, 0x7F, 0x9F, 0xA6, 0x9E, 0x94, 0x91, 0x92, 0x92, 0x93, 0x94, 0x93, 0x91, 0x91, 0x91, 0x92, 0x93, 0x93, 0x93, 0x94, 0x93, 0x93, 0x94, 0x93, 0x94, 0x94, 0x95, 0x9F, 0xA7, 0x9F, 0x7F, 0x53, 0x2A, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2A, 0x53, 0x81, 0xA1, 0xA8, 0xA0, 0x96, 0x93, 0x94, 0x94, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x93, 0x94, 0x93, 0x93, 0x95, 0x9F, 0xA8, 0xA1, 0x80, 0x53, 0x29, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2D, 0x59, 0x88, 0xAB, 0xB4, 0xAA, 0x9F, 0x9C, 0x9C, 0x9D, 0x9C, 0x9C, 0x9D, 0x9C, 0x9D, 0x9C, 0x9C, 0x9C, 0x9C, 0x9C, 0x9D, 0x9D, 0x9C, 0x9C, 0x9C, 0x9D, 0x9D, 0x9F, 0xAA, 0xB3, 0xAB, 0x88, 0x54, 0x27, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E, 0x2E, 0x5D, 0x8F, 0xB4, 0xBC, 0xB3, 0xA7, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA4, 0xA7, 0xB3, 0xBC, 0xB4, 0x8C, 0x57, 0x26, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2E, 0x59, 0x89, 0xAC, 0xB4, 0xAB, 0x9F, 0x9D, 0x9E, 0x9D, 0x9E, 0x9D, 0x9E, 0x9D, 0x9D, 0x9D, 0x9D, 0x9E, 0x9E, 0x9D, 0x9D, 0x9E, 0x9D, 0x9E, 0x9D, 0x9E, 0x9D, 0xA0, 0xAB, 0xB4, 0xAC, 0x89, 0x56, 0x27, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x26, 0x49, 0x70, 0x8C, 0x93, 0x8A, 0x82, 0x81, 0x81, 0x81, 0x80, 0x81, 0x80, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x80, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x82, 0x8B, 0x91, 0x8B, 0x71, 0x4E, 0x2E, 0x14, 0x0D, 0x0E, 0x0D, 0x0C, 0x0C, 0x0C, 0x0D, 0x0C, 0x0D, 0x0D, 0x0E, 0x0E, 0x0C, 0x0A, 0x06, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x1D, 0x35, 0x50, 0x64, 0x67, 0x62, 0x5E, 0x5B, 0x5C, 0x5C, 0x5B, 0x5B, 0x5B, 0x5C, 0x5B, 0x5C, 0x5B, 0x5C, 0x5C, 0x5C, 0x5C, 0x5B, 0x5C, 0x5B, 0x5B, 0x5B, 0x5B, 0x5C, 0x60, 0x61, 0x5C, 0x4E, 0x43, 0x36, 0x2E, 0x2E, 0x30, 0x2D, 0x2B, 0x2A, 0x2A, 0x29, 0x29, 0x2C, 0x30, 0x35, 0x35, 0x2C, 0x1E, 0x11, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0x13, 0x25, 0x37, 0x42, 0x43, 0x42, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3E, 0x3F, 0x3E, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3E, 0x3F, 0x3E, 0x3B, 0x38, 0x3A, 0x42, 0x4E, 0x5A, 0x5E, 0x5A, 0x54, 0x53, 0x52, 0x51, 0x52, 0x56, 0x61, 0x6A, 0x6A, 0x56, 0x38, 0x1D, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x10, 0x1D, 0x2C, 0x34, 0x34, 0x34, 0x34, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x34, 0x33, 0x32, 0x3A, 0x52, 0x71, 0x89, 0x90, 0x88, 0x80, 0x7F, 0x7D, 0x7D, 0x7E, 0x81, 0x8E, 0x98, 0x96, 0x78, 0x4E, 0x27, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1D, 0x2C, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x34, 0x45, 0x65, 0x8E, 0xAB, 0xB2, 0xAA, 0xA1, 0xA0, 0xA0, 0xA2, 0xA2, 0xA1, 0xA6, 0xA9, 0x9C, 0x7C, 0x51, 0x2A, 0x0D, 0x02, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1E, 0x2E, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x35, 0x3F, 0x59, 0x78, 0x98, 0xB0, 0xB5, 0xAE, 0xA7, 0xA8, 0xB0, 0xB7, 0xB6, 0xAC, 0xA0, 0x94, 0x7D, 0x61, 0x43, 0x28, 0x14, 0x0C, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1E, 0x2E, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x3A, 0x4B, 0x61, 0x77, 0x88, 0x96, 0xA4, 0xA4, 0xA0, 0x9E, 0xA2, 0xAE, 0xB8, 0xB8, 0xA5, 0x88, 0x6E, 0x54, 0x42, 0x34, 0x29, 0x1F, 0x18, 0x10, 0x09, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1D, 0x2D, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x38, 0x4E, 0x6F, 0x8C, 0x97, 0x96, 0x93, 0x95, 0x94, 0x93, 0x92, 0x96, 0xA3, 0xAD, 0xAA, 0x8F, 0x69, 0x4A, 0x38, 0x33, 0x31, 0x2F, 0x2D, 0x25, 0x19, 0x0C, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x10, 0x1E, 0x2C, 0x34, 0x34, 0x34, 0x33, 0x33, 0x32, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x34, 0x34, 0x34, 0x37, 0x49, 0x69, 0x8F, 0xAA, 0xAD, 0xA2, 0x96, 0x93, 0x93, 0x95, 0x94, 0x93, 0x97, 0x99, 0x8D, 0x71, 0x50, 0x39, 0x32, 0x32, 0x33, 0x32, 0x33, 0x30, 0x1F, 0x10, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07, 0x14, 0x25, 0x37, 0x42, 0x46, 0x46, 0x41, 0x3B, 0x35, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x34, 0x37, 0x3C, 0x44, 0x52, 0x6C, 0x88, 0xA5, 0xB8, 0xB8, 0xAD, 0xA0, 0x9C, 0x9F, 0xA3, 0xA2, 0x97, 0x8B, 0x7E, 0x68, 0x53, 0x3E, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x20, 0x10, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x1C, 0x35, 0x50, 0x64, 0x6F, 0x6F, 0x64, 0x52, 0x40, 0x35, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x34, 0x3D, 0x4E, 0x64, 0x7C, 0x91, 0x9F, 0xAC, 0xB6, 0xB5, 0xAE, 0xA6, 0xA5, 0xAC, 0xB2, 0xAD, 0x98, 0x7C, 0x60, 0x48, 0x3B, 0x35, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x30, 0x1F, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x26, 0x49, 0x71, 0x8D, 0x9D, 0x9E, 0x8D, 0x72, 0x50, 0x39, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x35, 0x41, 0x5C, 0x7F, 0x9B, 0xA6, 0xA4, 0xA0, 0xA1, 0xA2, 0x9F, 0x9E, 0x9F, 0xA8, 0xB1, 0xAA, 0x8D, 0x68, 0x48, 0x36, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2C, 0x1C, 0x0E, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2D, 0x59, 0x89, 0xAE, 0xC0, 0xBF, 0xAC, 0x8D, 0x67, 0x46, 0x36, 0x34, 0x34, 0x35, 0x35, 0x33, 0x31, 0x2F, 0x2D, 0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x39, 0x56, 0x7A, 0x94, 0x97, 0x8D, 0x81, 0x7E, 0x7D, 0x80, 0x80, 0x83, 0x8A, 0x92, 0x8B, 0x72, 0x50, 0x39, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2C, 0x1C, 0x0D, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2E, 0x5D, 0x8E, 0xB4, 0xC5, 0xC2, 0xB0, 0x97, 0x76, 0x5A, 0x45, 0x40, 0x46, 0x4B, 0x4A, 0x41, 0x34, 0x28, 0x1F, 0x1C, 0x1C, 0x1D, 0x1C, 0x1E, 0x28, 0x3D, 0x57, 0x69, 0x6A, 0x60, 0x56, 0x53, 0x54, 0x58, 0x5C, 0x5F, 0x63, 0x67, 0x62, 0x52, 0x3F, 0x34, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x1B, 0x0C, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2D, 0x59, 0x89, 0xAC, 0xB8, 0xB3, 0xA4, 0x95, 0x81, 0x6F, 0x60, 0x5F, 0x6B, 0x75, 0x75, 0x61, 0x42, 0x27, 0x14, 0x0E, 0x0E, 0x0E, 0x0E, 0x10, 0x16, 0x20, 0x2D, 0x35, 0x34, 0x30, 0x2C, 0x2A, 0x2F, 0x37, 0x40, 0x45, 0x45, 0x44, 0x41, 0x3B, 0x35, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x1A, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2B, 0x53, 0x81, 0xA0, 0xAA, 0xA1, 0x97, 0x91, 0x8B, 0x85, 0x82, 0x84, 0x91, 0x9C, 0x99, 0x7C, 0x51, 0x2A, 0x0D, 0x03, 0x02, 0x03, 0x03, 0x05, 0x08, 0x0A, 0x0C, 0x0E, 0x0E, 0x0E, 0x0C, 0x0D, 0x16, 0x23, 0x31, 0x36, 0x35, 0x35, 0x34, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x2B, 0x1A, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2A, 0x52, 0x7E, 0x9F, 0xA5, 0x9C, 0x92, 0x94, 0x98, 0x9D, 0x9F, 0x9F, 0xA3, 0xA6, 0x9A, 0x78, 0x4E, 0x28, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0B, 0x18, 0x27, 0x2E, 0x2F, 0x32, 0x32, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x2C, 0x1F, 0x11, 0x07, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2A, 0x52, 0x7F, 0x9F, 0xA5, 0x9A, 0x91, 0x95, 0xA1, 0xAD, 0xB3, 0xA9, 0x9C, 0x8C, 0x74, 0x56, 0x39, 0x1E, 0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x0F, 0x1A, 0x21, 0x27, 0x2D, 0x32, 0x33, 0x32, 0x32, 0x33, 0x32, 0x32, 0x32, 0x32, 0x33, 0x32, 0x32, 0x33, 0x2E, 0x24, 0x1A, 0x11, 0x0C, 0x07, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2A, 0x53, 0x7F, 0x9E, 0xA4, 0x9A, 0x92, 0x97, 0xA6, 0xB5, 0xB8, 0xA2, 0x82, 0x63, 0x42, 0x2C, 0x1E, 0x11, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x08, 0x0D, 0x12, 0x1B, 0x26, 0x2F, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x32, 0x33, 0x32, 0x32, 0x32, 0x31, 0x2A, 0x24, 0x1F, 0x18, 0x0E, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2A, 0x52, 0x7F, 0x9F, 0xA5, 0x9C, 0x92, 0x94, 0xA2, 0xAF, 0xAB, 0x8D, 0x62, 0x39, 0x19, 0x0C, 0x09, 0x06, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x02, 0x08, 0x12, 0x1E, 0x2D, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x32, 0x33, 0x32, 0x30, 0x2F, 0x2D, 0x25, 0x17, 0x0A, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2A, 0x53, 0x81, 0xA0, 0xAA, 0xA1, 0x97, 0x94, 0x99, 0x9C, 0x8F, 0x6F, 0x44, 0x1F, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x1B, 0x2B, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x31, 0x21, 0x13, 0x07, 0x02, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2C, 0x58, 0x89, 0xAC, 0xB9, 0xB3, 0xA4, 0x97, 0x8C, 0x81, 0x6B, 0x4E, 0x2F, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x1C, 0x2F, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x29, 0x1C, 0x12, 0x0D, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x2F, 0x5D, 0x8F, 0xB4, 0xC6, 0xC2, 0xAF, 0x98, 0x7E, 0x64, 0x4B, 0x36, 0x23, 0x10, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0D, 0x1F, 0x31, 0x32, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2D, 0x25, 0x1F, 0x19, 0x10, 0x09, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x2D, 0x59, 0x89, 0xAE, 0xC0, 0xBF, 0xAC, 0x8E, 0x6A, 0x4D, 0x38, 0x2D, 0x1F, 0x10, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0F, 0x1E, 0x2F, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x32, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x30, 0x2E, 0x2D, 0x26, 0x18, 0x0C, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x26, 0x49, 0x70, 0x8E, 0x9E, 0x9E, 0x8E, 0x72, 0x52, 0x3D, 0x34, 0x2E, 0x1E, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0C, 0x19, 0x26, 0x2D, 0x2F, 0x31, 0x32, 0x32, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x30, 0x1F, 0x10, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x1C, 0x34, 0x50, 0x64, 0x6E, 0x6F, 0x63, 0x52, 0x41, 0x37, 0x33, 0x2E, 0x1E, 0x0D, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x09, 0x11, 0x19, 0x20, 0x26, 0x2D, 0x33, 0x32, 0x32, 0x33, 0x32, 0x33, 0x32, 0x32, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x22, 0x11, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x14, 0x25, 0x37, 0x42, 0x46, 0x46, 0x41, 0x3C, 0x37, 0x34, 0x33, 0x2C, 0x1B, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, 0x08, 0x0D, 0x12, 0x1C, 0x28, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x23, 0x11, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x0F, 0x1D, 0x2C, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x33, 0x32, 0x2B, 0x1B, 0x0C, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x08, 0x13, 0x21, 0x31, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x32, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x30, 0x1F, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1D, 0x2D, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2D, 0x1E, 0x11, 0x07, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x17, 0x25, 0x2B, 0x2C, 0x2C, 0x2C, 0x2C, 0x2B, 0x2B, 0x2C, 0x2B, 0x2B, 0x2C, 0x2C, 0x2B, 0x2C, 0x2C, 0x2C, 0x26, 0x18, 0x0C, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x10, 0x20, 0x31, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x2E, 0x24, 0x19, 0x12, 0x0C, 0x09, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x0E, 0x18, 0x1D, 0x1D, 0x1C, 0x1C, 0x1C, 0x1D, 0x1C, 0x1D, 0x1C, 0x1C, 0x1D, 0x1C, 0x1D, 0x1D, 0x1C, 0x1D, 0x18, 0x10, 0x09, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x10, 0x21, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x31, 0x2A, 0x24, 0x1E, 0x19, 0x11, 0x08, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x08, 0x0C, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0D, 0x08, 0x03, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x0F, 0x1F, 0x30, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x32, 0x31, 0x2F, 0x2C, 0x25, 0x19, 0x0C, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x02, 0x03, 0x02, 0x03, 0x03, 0x02, 0x03, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x03, 0x02, 0x03, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0D, 0x18, 0x26, 0x2C, 0x2F, 0x30, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x30, 0x1F, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x09, 0x10, 0x19, 0x1E, 0x24, 0x2A, 0x30, 0x33, 0x32, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x20, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x08, 0x0C, 0x12, 0x1A, 0x24, 0x2E, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x33, 0x30, 0x1E, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x02, 0x07, 0x11, 0x1E, 0x2C, 0x33, 0x33, 0x32, 0x33, 0x32, 0x32, 0x33, 0x33, 0x2D, 0x1D, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x0C, 0x1B, 0x2B, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x32, 0x2B, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0B, 0x1A, 0x2B, 0x33, 0x32, 0x33, 0x33, 0x32, 0x32, 0x33, 0x32, 0x2C, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0x1B, 0x2B, 0x33, 0x33, 0x33, 0x32, 0x32, 0x33, 0x33, 0x32, 0x2C, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1C, 0x2C, 0x32, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x1C, 0x0E, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1D, 0x2C, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2C, 0x1D, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1D, 0x2B, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x33, 0x2B, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1C, 0x2B, 0x32, 0x33, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x2B, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0E, 0x1C, 0x2B, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x2C, 0x1C, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0E, 0x1D, 0x2C, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x2C, 0x1D, 0x0E, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1F, 0x30, 0x33, 0x32, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x30, 0x1F, 0x10, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x10, 0x20, 0x32, 0x33, 0x33, 0x33, 0x33, 0x33, 0x32, 0x33, 0x33, 0x32, 0x20, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x0F, 0x1F, 0x30, 0x33, 0x33, 0x33, 0x32, 0x33, 0x32, 0x33, 0x32, 0x30, 0x1F, 0x0F, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x0D, 0x18, 0x26, 0x2C, 0x2C, 0x2C, 0x2B, 0x2C, 0x2C, 0x2C, 0x2C, 0x26, 0x18, 0x0D, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x08, 0x10, 0x19, 0x1D, 0x1D, 0x1D, 0x1D, 0x1C, 0x1C, 0x1C, 0x1C, 0x19, 0x10, 0x08, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x08, 0x0D, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0E, 0x0C, 0x08, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
        };

        if(false)//width > BMP_W && height > BMP_H)
        {
            uint08 *dst = pBuffer[previousBuffer].pPixels;

            int bmpX = mouseX - (BMP_W/2);
            int bmpY = mouseY - (BMP_H/2);

            if(bmpX < 0) { bmpX = 0; }
            if(bmpY < 0) { bmpY = 0; }
            if(bmpX > (width - BMP_W)) { bmpX = width - BMP_W; }
            if(bmpY > (height - BMP_H)) { bmpY = height - BMP_H; }

            dst += bmpY*width + bmpX;

            for(int y=0;y<BMP_H;y++)
            {
                for(int x=0;x<BMP_W;x++)
                {                
                    if(pMegaMan[y*BMP_W+x] > 20) { assignFunc(dst, pMegaMan[y*BMP_W+x]); }

                    dst++;
                }

                dst += width - BMP_W;
            }
        }
        //*/
    }

    AutoMutex threadSafe(pRenderFXMutex);

    RenderFXNode *pCur = pRenderFXNodeH;

    while(pCur != 0)
    {
        pCur->pRenderFX->Render();
        pCur->pRenderFX->Step();

        pCur = pCur->pNext;
    }

    return true;
}

bool Sirenic::Transform()
{
    if(!bCreated) { return false; }

    AutoMutex activeBufferLock(pBuffer[activeBuffer].pMutex);

    int waveX = 0, waveY = 0;

    //
    // calculate wave effect
    // @todo expose this externally
    //*
    {
        static float fWaveX = 0;
        static float fWaveXs = 0.024f;
        static float fWaveY = 0;
        static float fWaveYs = 0.040f;

        if(fWaveX > 3.1415f) { fWaveX = -3.1415f; }
        if(fWaveY > 3.1415f) { fWaveY = -3.1415f; }

        fWaveX += fWaveXs;
        fWaveY += fWaveYs;

        waveX = (int)(sinf(fWaveX)*5.0f);
        waveY = (int)(cosf(fWaveY)*5.0f);
    }
    //*/

    waveX = waveY = 0;

    //
    // transform
    //

    if(true)
    {
        uint08 *src = pBuffer[previousBuffer].pPixels;
        uint08 *dst = pBuffer[activeBuffer].pPixels;

        // sanity check (temporary)
        if(pitch != width)
            return false;

        uint08 *src0 = src - pitch;
        uint08 *src1 = src - 1;
        uint08 *src2 = src + 1;
        uint08 *src3 = src + pitch;

        uint32 strt = 0;
        uint32 stop = height*width;

        if(waveY > 0) { stop -= waveY*pitch; }
        if(waveX > 0) { stop -= waveX; }
        if(waveY < 0) { strt -= waveY*pitch; }
        if(waveX < 0) { strt -= waveX; }

        //
        // clean up edges caused by wave
        //
        /* Redundant?
        {
            int waveOffs = waveX + waveY*pitch;

            if(waveOffs > 0)
            {
                memset(dst, 0, waveOffs);
            }
            else
            {
                memset(dst + height*width + waveOffs - 1, 0, 0-waveOffs);
            }
        }
        //*/

        dst += waveX + waveY*pitch;

        int circOffs = rand()%(height*width);

        if(!bExtraSrcWeight)
        {
            for(uint d=strt;d<stop;d++)
            {
                uint32 o = pDeltaFieldFront[d].o;

                // transform with bilinear interpolation
                dst[d] = (uint08)
                (
                    (
                        (uint32)src0[o] * (uint32)pDeltaFieldFront[d].w0 +
                        (uint32)src1[o] * (uint32)pDeltaFieldFront[d].w1 +
                        (uint32)src2[o] * (uint32)pDeltaFieldFront[d].w2 +
                        (uint32)src3[o] * (uint32)pDeltaFieldFront[d].w3
                    ) >> 8
                );

                int curOffs = (d+circOffs);

                if(curOffs >= height*width) { curOffs -= height*width; }

                int sub = pDitherMap[curOffs];

                if( (sub > 0) && (sub > dst[d]) ) { sub = dst[d]; }
                if( (sub < 0) && (dst[d] - sub > 255) ) { sub = - (255 - dst[d]); }

                dst[d] -= sub;
            }
        }
        else
        {
            for(uint d=strt;d<stop;d++)
            {
                uint32 o = pDeltaFieldFront[d].o;

                /*! transform with bilinear interpolation with extra source weight */
                dst[d] = (uint08)
                (
                    (
                        (uint32)src0[o] * (uint32)pDeltaFieldFront[d].w0 +
                        (uint32)src1[o] * (uint32)pDeltaFieldFront[d].w1 +
                        (uint32)src2[o] * (uint32)pDeltaFieldFront[d].w2 +
                        (uint32)src3[o] * (uint32)pDeltaFieldFront[d].w3 +
                        src[d]*128
                    ) / 384
                );

                int curOffs = (d+circOffs);

                if(curOffs >= height*width) { curOffs -= height*width; }

                int sub = pDitherMap[curOffs];

                if( (sub > 0) && (sub > dst[d]) ) { sub = dst[d]; }
                if( (sub < 0) && (dst[d] - sub > 255) ) { sub = - (255 - dst[d]); }

                dst[d] -= sub;
            }
        }

        /*! Render Delta Field */
        if(false)
        {
            #define DELTA_FIELD_RESOLUTION 8

            int v=0;

            for(int y=0;y<height;y += DELTA_FIELD_RESOLUTION, v += (DELTA_FIELD_RESOLUTION-1)*width)
            {
                for(int x=0;x<width;x += DELTA_FIELD_RESOLUTION, v += DELTA_FIELD_RESOLUTION)
                {
                    int dy = pDeltaFieldFront[v].o / width;
                    int dx = pDeltaFieldFront[v].o % width;

                    if(dx < 0) { dx = 0; }
                    if(dx >= width) { dx = 0; }
                    if(dy < 0) { dy = 0; }
                    if(dy >= height) { dy = 0; }

                    //dx = 0;

                    roDrawLine(src, pitch, x, y, dx, dy, 255, assignFunc);
                }
            }
        }
    }
    else
    {
        uint08 *dst = pBuffer[activeBuffer].pPixels;

        memset(dst, 0, height*pitch);
    }

    //
    // render waveform here...
    //

    if(bWaveformEnabled)
    {
        AutoMutex threadSafe(pWaveformMutex);

        sint08 *src = (sint08*)pWaveform[activeWaveform];
        uint08 *dst = pBuffer[activeBuffer].pPixels;

        float sx = 0.0;
        float sxs = (float)width / 576.0f;

        float lsx = 0;
        float lsy = (float)height / 2.0f;

        for(uint x=0;x<576;x++)
        {
            float sy = (float)height/2.0f + (float)src[x]/128.0f*(float)height*waveformScale;

            sx += sxs;

            //
            // clamp
            //

            if(sx < 0) { sx = 0.0f; }

            if(sy < 0) { sy = 0.0f; }

            /*! @todo -1 and -2.0f are hacks to keep this in bound..verify roDrawLineWu for bounds checking */
            if(sx >= width-1) { sx = (float)width - 2.0f; }

            if(sy >= height-1) { sy = (float)height - 2.0f; }

            //
            // line rendering
            // 

            Sirenic::roDrawLineWu(dst, pitch, lsx, lsy, sx, sy, waveformColor);

            //
            // dot rendering
            //

            // Sirenic::roDrawPixelWu(dst, pitch, sx, sy, 1.0f);

            lsy = sy;
            lsx = sx;
        }
    }

    //
    // darken center
    //
    //*
    if(darkenCenter)
    {
        uint08 *dst = pBuffer[activeBuffer].pPixels;

        int rx = (int)(0.5f * width + waveX);
        int ry = (int)(0.5f * height + waveY);

        int ind = ry*pitch + rx;

        dst[ind] -= 5;

        if(dst[ind] > 250)
        {
            dst[ind] = 0;
        }
    }
    //*/

    //
    // darken edges
    //

    if(darkenEdges)
    {
        int x, y;

        uint08 *dst = pBuffer[activeBuffer].pPixels;

        //
        // darken left/right edges
        //

        {
            uint08 *dstL = dst;
            uint08 *dstR = dst + width;

            int lenL = 1;
            int lenR = 1;

            if(waveX > 0)
            {
                lenL += waveX - 1;
            }
            else if(waveX < 0)
            {
                lenR += (0 - waveX) - 1;
            }

            dstR -= lenR;

        for(y=0;y<height;y++)
        {
                for(x=0;x<lenL;x++) { dstL[x] = 0; }
                for(x=0;x<lenR;x++) { dstR[x] = 0; }

                dstL += pitch;
                dstR += pitch;
        }
        }

        //
        // darken top/bottom edges
        //

        {
            uint08 *dstT = dst;
            uint08 *dstB = dst + (height - 1)*pitch;

            memset(dstT, 0, width);
            memset(dstB, 0, width);

            int lenT = 1;
            int lenB = 1;

            if(waveY > 0)
            {
                lenT += waveY - 1;
            }
            else if(waveY < 0)
            {
                lenB += (0 - waveY) - 1;
            }

            for(y=0;y<lenT;y++)
            {
                memset(dstT, 0, width);

                dstT += pitch;
            }

            dstB -= pitch*lenB;

            for(y=0;y<lenB;y++)
            {
                memset(dstB, 0, width);

                dstB += pitch;
            }
        }
    }

    // advance active buffer
    {
        AutoMutex bufferLock(pBufferMutex);

        previousBuffer = activeBuffer;

        activeBuffer++;

        if(activeBuffer >= bufferCount)
        {
            activeBuffer = 0;
        }
    }

    return true;
}

void Sirenic::UpdateWaveform(uint08 *pNewWaveform)
{
    AutoMutex threadSafe(pWaveformMutex);

    // copy into waveform backbuffer
    memcpy(pWaveform[1 - activeWaveform], pNewWaveform, 576);

    activeWaveform = 1 - activeWaveform;

    return;
}

void Sirenic::UpdateWaveformScale(float waveformScale)
{
    AutoMutex threadSafe(pWaveformMutex);

    this->waveformScale = waveformScale;

    return;
}

uint08 *Sirenic::LockViewBuffer()
{
    if(!bCreated) { return 0; }

    pViewMutex->Wait();

    pBuffer[viewBuffer].pMutex->Wait();

    return pBuffer[viewBuffer].pPixels;
}

void Sirenic::UnlockViewBuffer()
{
    if(!bCreated) { return; }

    pBuffer[viewBuffer].pMutex->Release();

    pViewMutex->Release();

    return;
}

uint08 *Sirenic::GetViewBuffer()
{
    if(!bCreated) { return 0; }

    return pBuffer[viewBuffer].pPixels;
}

void Sirenic::AdvanceViewBuffer()
{
    if(!bCreated) { return; }

    AutoMutex threadSafe(pViewMutex);

    viewBuffer++;

    if(viewBuffer >= bufferCount)
    {
        viewBuffer = 0;
    }

    return;
}
